# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
from __future__ import annotations

import re

import pytest
from numpy import array
from numpy import inf
from numpy import nan_to_num
from numpy.testing import assert_almost_equal

from gemseo_mlearning.active_learning.acquisition_criteria.level_set.ef import EF
from gemseo_mlearning.active_learning.acquisition_criteria.level_set.ei import EI
from gemseo_mlearning.active_learning.acquisition_criteria.level_set.u import U


@pytest.mark.parametrize(
    ("cls", "input_value", "expected"),
    [
        (U, array([0.25]), array([0.3535534])),
        (EF, array([0.25]), array([0.5547019])),
        (EI, array([0.25]), array([0.6640591])),
        (U, array([[0.25]] * 2), array([[0.3535534]] * 2)),
        (EF, array([[0.25]] * 2), array([[0.5547019]] * 2)),
        (EI, array([[0.25]] * 2), array([[0.6640591]] * 2)),
    ],
)
def test_level_set_kriging(kriging_distribution, cls, input_value, expected):
    """Check the criteria deriving from BaseLevelSet with GP regressor."""
    criterion = cls(kriging_distribution, 0.5)
    assert_almost_equal(criterion.func(input_value), expected)
    assert criterion._mc_size == 10000
    assert criterion._batch_size == 1


@pytest.mark.parametrize(
    ("cls", "input_value", "expected"),
    [
        (U, array([0.25]), array([0.35])),
        (EF, array([0.25]), array([0.41])),
        (EI, array([0.25]), array([1])),
        (U, array([[0.25]] * 2), array([[0.35]] * 2)),
        (EF, array([[0.25]] * 2), array([[0.41]] * 2)),
        (EI, array([[0.25]] * 2), array([[1]] * 2)),
    ],
)
def test_level_set_regressor(algo_distribution, cls, input_value, expected):
    """Check the criteria deriving from BaseLevelSet with non GP regressor."""
    criterion = cls(algo_distribution, 0.5)
    assert_almost_equal(criterion.func(input_value), expected, decimal=2)
    assert criterion._mc_size == 10000
    assert criterion._batch_size == 1


@pytest.mark.parametrize(
    ("cls", "expected", "kwargs"),
    [
        (U, array([8.1571496e-08]), {}),
        (
            EF,
            0.9428005,
            {},
        ),
        (EI, array([0.8888888]), {}),
        (
            U,
            array([8.1571496e-08]),
            {"mc_size": 20000},
        ),
        (
            EF,
            0.942708,
            {"mc_size": 20000},
        ),
        (
            EI,
            array([0.8888888]),
            {"mc_size": 20000},
        ),
    ],
)
def test_level_set_parallel(kriging_distribution, cls, expected, kwargs):
    """Check the parallelized criteria deriving from BaseLevelSet."""
    criterion = cls(
        regressor_distribution=kriging_distribution,
        output_value=0.5,
        batch_size=2,
        **kwargs,
    )
    expected_mc_size = kwargs.get("mc_size", 10000)
    input_value = array([[0.25], [0.25]])
    assert_almost_equal(criterion.func(input_value), expected)
    assert criterion._mc_size == expected_mc_size
    assert criterion._batch_size == 2


@pytest.mark.parametrize("cls", [EF, EI])
@pytest.mark.parametrize("distribution", ["algo_distribution", "kriging_distribution"])
def test_improvement_at_training_point(cls, distribution, request):
    """Check that the improvement criteria take predefined values a training point."""
    distribution = request.getfixturevalue(distribution)
    criterion = cls(distribution, 1)
    criterion._compute_standard_deviation = lambda x: array([0])
    assert_almost_equal(criterion.evaluate(array([0.0])), array([0.0]))


@pytest.mark.parametrize(
    ("distribution"),
    [
        ("kriging_distribution"),
        ("algo_distribution"),
    ],
)
def test_u_at_training_point(distribution, request):
    """Check that the U criterion at a training point is infinity."""
    distribution = request.getfixturevalue(distribution)
    u = U(distribution, 1)
    u._compute_standard_deviation = lambda x: array([0])
    infinity = nan_to_num(inf)

    # Case where mean=mu(x) and std=0 => U = abs(1-mu(x))/0
    assert_almost_equal(u.evaluate(array([0.0])), array([infinity]))

    # Case where mean=1 and std=0 => U = abs(1-1)/0 = 0/0
    u._compute_mean = lambda x: array([1])
    assert_almost_equal(u.evaluate(array([0.0])), array([infinity]))


@pytest.mark.parametrize("cls", [EF, EI])
def test_improvement_parallel_at_training_point(cls, kriging_distribution):
    """Check that the improvement criteria take predefined values a training point."""
    criterion = cls(kriging_distribution, 1, batch_size=2)
    criterion._compute_samples = lambda x: TypeError
    assert_almost_equal(criterion.evaluate(array([0.0, 0.0])), array([0.0]))


@pytest.mark.parametrize("std", [array([0]), TypeError])
def test_u_parallel_at_training_point(kriging_distribution, std):
    """Check that the discrepancy criterion take predefined values a training point."""
    u = U(kriging_distribution, 1, batch_size=2)
    u._compute_variance = lambda x: std
    assert_almost_equal(u.evaluate(array([0.0, 0.0])), nan_to_num(array([inf])))


@pytest.mark.parametrize("cls", [U, EI, EF])
def test_bad_parallel_regressor(algo_distribution, cls):
    """Check that parallelized criteria with non GP regressor lead to failure."""
    criterion = cls(algo_distribution, batch_size=2, output_value=1)

    with pytest.raises(
        NotImplementedError,
        match=re.escape(
            "Parallelization with batch_size > 1 is not yet implemented "
            "for regressors that are not based on a random process."
        ),
    ):
        criterion.func(array([0.123]))



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""The Rosenbrock function and its gradient."""

from __future__ import annotations

from typing import Final

from numpy import array
from numpy import ndarray

__A: Final[float] = 100.0


def compute_output(x: ndarray) -> float:
    """Compute the output of the Rosenbrock function.

    Args:
        x: The input value.

    Returns:
        The output value.
    """
    return (1 - x[0]) ** 2 + __A * (x[1] - x[0] ** 2) ** 2


def compute_gradient(x: ndarray) -> ndarray:
    """Compute the gradient of the Rosenbrock function.

    Args:
        x: The input value.

    Returns:
        The value of the gradient of the Rosenbrock function.
    """
    return array([
        4 * __A * (x[0] ** 3 - x[0] * x[1]) + 2 * (x[0] - 1),
        2 * __A * (x[1] - x[0] ** 2),
    ])



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from gemseo import sample_disciplines
from gemseo.algos.design_space import DesignSpace
from gemseo.disciplines.analytic import AnalyticDiscipline
from gemseo.mlearning.regression.algos.gpr import GaussianProcessRegressor

from gemseo_mlearning.active_learning.active_learning_algo import ActiveLearningAlgo

if TYPE_CHECKING:
    from gemseo.datasets.io_dataset import IODataset


@pytest.fixture(scope="module")
def discipline() -> AnalyticDiscipline:
    """The original discipline."""
    return AnalyticDiscipline({"z": "(1-x)**2+100*(y-x**2)**2"}, name="Rosenbrock")


@pytest.fixture(scope="module")
def input_space() -> DesignSpace:
    """The input space of the surrogate model."""
    design_space = DesignSpace()
    design_space.add_variable("x", lower_bound=-2, upper_bound=2, value=1.0)
    design_space.add_variable("y", lower_bound=-2, upper_bound=2, value=1.0)
    return design_space


@pytest.fixture(scope="module")
def learning_dataset(discipline, input_space) -> IODataset:
    """The learning dataset."""
    return sample_disciplines(
        [discipline], input_space, "z", algo_name="OT_OPT_LHS", n_samples=30
    )


@pytest.fixture(scope="module")
def active_learning_algo(
    input_space, learning_dataset, discipline
) -> ActiveLearningAlgo:
    """The active learning algorithm."""
    regressor = GaussianProcessRegressor(learning_dataset)
    algo = ActiveLearningAlgo("Minimum", input_space, regressor)
    algo.acquire_new_points(discipline)
    return algo



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""A factory of acquisition criteria."""

from __future__ import annotations

from abc import abstractmethod
from typing import ClassVar

from gemseo.core.base_factory import BaseFactory

from gemseo_mlearning.active_learning.acquisition_criteria.base_acquisition_criterion import (  # noqa: E501
    BaseAcquisitionCriterion,
)


class BaseAcquisitionCriterionFactory(BaseFactory):
    """A factory of acquisition criteria."""

    _CLASS = BaseAcquisitionCriterion
    _PACKAGE_NAMES = ("gemseo_mlearning.active_learning.acquisition_criteria",)

    @property
    @abstractmethod
    def _DEFAULT_CLASS_NAME(self) -> str:  # noqa: N802
        """The name of the default class."""

    def get_class(self, name: str = "") -> type[BaseAcquisitionCriterion]:  # noqa: D102
        return super().get_class(name or self._DEFAULT_CLASS_NAME)


class BaseAcquisitionCriterionFamilyFactory:
    """The factory of families of acquisition criteria."""

    ACQUISITION_CRITERION_FAMILY_FACTORY: ClassVar[BaseAcquisitionCriterionFactory]



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
from __future__ import annotations

from operator import add
from operator import mul
from operator import sub
from operator import truediv

import pytest
from gemseo.core.mdo_functions.mdo_function import MDOFunction
from gemseo.mlearning.regression.algos.linreg import LinearRegressor
from numpy import array

from gemseo_mlearning.active_learning.acquisition_criteria.exploration.distance import (
    Distance,
)
from gemseo_mlearning.active_learning.acquisition_criteria.minimum.ei import EI
from gemseo_mlearning.active_learning.distributions.regressor_distribution import (
    RegressorDistribution,
)


@pytest.fixture(scope="module")
def algo_distribution(dataset) -> RegressorDistribution:
    """The distribution of the linear regression model."""
    distribution = RegressorDistribution(LinearRegressor(dataset), size=3)
    distribution.learn()
    return distribution


@pytest.fixture(scope="module")
def criterion_1(algo_distribution):
    """First data acquisition criterion to check operations."""
    return EI(algo_distribution)


@pytest.fixture(scope="module")
def criterion_2(algo_distribution):
    """Second data acquisition criterion to check operations."""
    return Distance(algo_distribution)


@pytest.mark.parametrize("operator", [add, mul, sub, truediv])
@pytest.mark.parametrize("second_operand", [2.0, None])
def test_operations(criterion_1, criterion_2, operator, second_operand):
    """Check elementary operations with two MLDataAcquisitionCriterion."""
    x = array([0.5])
    c_1 = criterion_1.func(x)
    if second_operand is None:
        c_2 = criterion_2.func(x)
        criterion_2 = criterion_2
    else:
        c_2 = second_operand
        criterion_2 = second_operand

    criterion_3 = operator(criterion_1, criterion_2)
    assert criterion_3.func(x) == operator(c_1, c_2)


def test_neg(criterion_1):
    """Check the neg operator."""
    x = array([0.5])
    neg_criterion_1 = -criterion_1
    assert neg_criterion_1.func(x) == -criterion_1.func(x)


def test_linear_combination(algo_distribution):
    """Check that a combination of MLDataAcquisitionCriterion is an MDOFunction."""
    criterion_1 = EI(algo_distribution)
    criterion_2 = Distance(algo_distribution)
    criterion_3 = criterion_1 * 0.2 + criterion_2 * 0.8
    assert isinstance(criterion_3, MDOFunction)
    x_new = array([0.5])
    assert (
        criterion_3.func(x_new)
        == criterion_1.func(x_new) * 0.2 + criterion_2.func(x_new) * 0.8
    )



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Tests for the surrogate-based optimizer."""

from __future__ import annotations

from unittest.mock import MagicMock

import pytest
from gemseo.algos.design_space import DesignSpace
from gemseo.algos.optimization_problem import OptimizationProblem
from gemseo.core.mdo_functions.mdo_function import MDOFunction
from gemseo.problems.optimization.rastrigin import Rastrigin
from numpy import array
from pandas._testing import assert_frame_equal

from gemseo_mlearning.algos.opt.core.surrogate_based_optimizer import (
    SurrogateBasedOptimizer,
)
from gemseo_mlearning.algos.opt.sbo_settings import SBO_Settings
from gemseo_mlearning.algos.opt.sbo_settings import SBOSettings


@pytest.mark.parametrize(
    ("regression_algorithm", "regression_settings"),
    [
        ("GaussianProcessRegressor", {}),
        ("OTGaussianProcessRegressor", {"use_hmat": False}),
    ],
)
def test_all_acquisitions_made(regression_algorithm, regression_settings):
    """Check the execution of the surrogate-based optimizer with all acquisitions."""
    assert (
        SurrogateBasedOptimizer(
            Rastrigin(),
            "PYDOE_FULLFACT",
            5,
            regression_algorithm=regression_algorithm,
            regression_settings=regression_settings,
            n_samples=10,
        ).execute(1)
        == "All the data acquisitions have been made."
    )


def test_known_acquired_input_data():
    """Check the termination when the acquired input data is already known."""
    space = DesignSpace()
    space.add_variable("x", lower_bound=0, upper_bound=1)
    problem = OptimizationProblem(space)
    problem.objective = MDOFunction(lambda _: 0, "f")
    assert (
        SurrogateBasedOptimizer(
            problem,
            "CustomDOE",
            2,
            regression_algorithm="LinearRegressor",
            samples=array([[0.0]]),
        ).execute(2)
        == "The acquired input data is already known."
    )


def test_convergence_on_rastrigin():
    """Check the surrogate-based optimizer on Rastrigin's function."""

    def listener(x):
        return

    problem = Rastrigin()
    problem.database.add_store_listener(listener)
    problem.database.add_store_listener = MagicMock()
    SurrogateBasedOptimizer(
        problem,
        "DIFFERENTIAL_EVOLUTION",
        doe_size=20,
        max_iter=1000,
        popsize=50,
        seed=1,
    ).execute(5)
    assert problem.optimum.objective < 0.12

    # Check that the optimizer resets the listener after the sub-algo has removed it.
    problem.database.add_store_listener.assert_called()


def test_stratified_algorithm():
    """Check the use of a stratified algorithm for the initial sampling."""
    assert (
        SurrogateBasedOptimizer(
            Rastrigin(),
            "DIFFERENTIAL_EVOLUTION",
            doe_algorithm="OT_AXIAL",
            doe_settings={"centers": [0.5, 0.5], "levels": [0.1, 0.2]},
            max_iter=10,
        ).execute(1)
        == "All the data acquisitions have been made."
    )


def test_ml_regression_algo_instance(regression_algorithm):
    """Check the execution of the surrogate-based optimizer with an
    BaseMLRegressionAlgo.
    """
    optimizer = SurrogateBasedOptimizer(
        Rastrigin(),
        "CustomDOE",
        regression_algorithm=regression_algorithm,
        samples=array([[0.03, 0.03]]),
    )
    optimizer.execute(1)
    dataset = optimizer._SurrogateBasedOptimizer__dataset

    optimizer = SurrogateBasedOptimizer(
        Rastrigin(),
        "CustomDOE",
        doe_size=5,
        doe_algorithm="OT_SOBOL",
        regression_algorithm="OTGaussianProcessRegressor",
        samples=array([[0.03, 0.03]]),
    )
    optimizer.execute(1)
    assert_frame_equal(optimizer._SurrogateBasedOptimizer__dataset, dataset)


def test_alias():
    """Verify that SBOSettings is an alias of SBO_Settings."""
    assert SBOSettings == SBO_Settings



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Active learning algorithm."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING
from typing import Any
from typing import ClassVar

from gemseo.algos._progress_bars.custom_tqdm_progress_bar import LOGGER as TQDM_LOGGER
from gemseo.algos._progress_bars.custom_tqdm_progress_bar import CustomTqdmProgressBar
from gemseo.algos.database import Database
from gemseo.algos.design_space import DesignSpace
from gemseo.algos.doe.factory import DOELibraryFactory
from gemseo.algos.opt.factory import OptimizationLibraryFactory
from gemseo.algos.optimization_problem import OptimizationProblem
from gemseo.datasets.io_dataset import IODataset
from gemseo.mlearning.regression.algos.base_random_process_regressor import (
    BaseRandomProcessRegressor,
)
from gemseo.mlearning.regression.algos.base_regressor import BaseRegressor
from gemseo.utils.logging_tools import LoggingContext
from gemseo.utils.logging_tools import OneLineLogging
from numpy import array
from numpy import hstack
from numpy import newaxis
from numpy import tile
from pandas import concat

from gemseo_mlearning.active_learning.acquisition_criteria.base_acquisition_criterion_family import (  # noqa: E501
    AcquisitionCriterionFamilyFactory,
)
from gemseo_mlearning.active_learning.distributions import KrigingDistribution
from gemseo_mlearning.active_learning.distributions import RegressorDistribution
from gemseo_mlearning.active_learning.visualization.acquisition_view import (
    AcquisitionView,
)
from gemseo_mlearning.active_learning.visualization.qoi_history_view import (
    QOIHistoryView,
)

if TYPE_CHECKING:
    from pathlib import Path

    from gemseo.algos.base_driver_library import BaseDriverLibrary
    from gemseo.core.discipline.discipline import Discipline
    from gemseo.mlearning.core.algos.ml_algo import DataType
    from gemseo.post.dataset.lines import Lines
    from matplotlib.figure import Figure

    from gemseo_mlearning.active_learning.acquisition_criteria.base_acquisition_criterion import (  # noqa: E501
        BaseAcquisitionCriterion,
    )
    from gemseo_mlearning.active_learning.distributions.base_regressor_distribution import (  # noqa: E501
        BaseRegressorDistribution,
    )

LOGGER = logging.getLogger(__name__)


class ActiveLearningAlgo:
    """An active learning algorithm using a regressor and acquisition criteria."""

    __acquisition_algo: BaseDriverLibrary
    """The algorithm to find the new training point(s)."""

    __acquisition_algo_settings: dict[str, Any]
    """The settings of the algorithm to find the new training points()."""

    __acquisition_criterion: BaseAcquisitionCriterion
    """The acquisition criterion."""

    __acquisition_problem: OptimizationProblem
    """The acquisition problem."""

    __acquisition_view: AcquisitionView | None
    """View of points acquired during active learning if the input dimension is 2."""

    __database: Database
    """The concatenation of the optimization histories related to the new points."""

    __input_space: DesignSpace
    """The input space on which to look for the new learning point."""

    __default_algo_name: ClassVar[str] = "MultiStart"
    """The name of the default algorithm to find the new training point(s).

    Typically a DoE or an optimizer.
    """

    __default_algo_settings: ClassVar[dict[str, Any]] = {
        "max_iter": 200,
        "n_start": 20,
        "opt_algo_name": "SLSQP",
    }
    """The names and values of the default algorithm settings."""

    __distribution: BaseRegressorDistribution
    """The distribution of the machine learning algorithm."""

    __n_initial_samples: int
    """The number of initial samples."""

    __batch_size: int
    """The number of points to be acquired in parallel.

    If `1`, acquire points sequentially.
    """

    __n_evaluations_history: list[int]
    """The history of the number of evaluations."""

    __qoi_history: list[float]
    """The history of the quantity of interest."""

    def __init__(
        self,
        criterion_family_name: str,
        input_space: DesignSpace,
        regressor: BaseRegressor | BaseRegressorDistribution,
        criterion_name: str = "",
        batch_size: int = 1,
        mc_size: int = 10_000,
        **criterion_arguments: Any,
    ) -> None:
        """
        Args:
            criterion_family_name: The name of a family of acquisition criteria,
                *e.g.* `"Minimum"`, `"Maximum"`, `"LevelSet"`, `"Quantile"`
                or `"Exploration"`.
            input_space: The input space on which to look for the new learning point.
            regressor: Either a regressor or a regressor distribution.
            criterion_name: The name of the acquisition criterion.
                If empty,
                use the default criterion of the family `criterion_family_name`.
            batch_size: The number of points to be acquired in parallel;
                if `1`, acquire points sequentially.
            mc_size: The sample size to estimate the acquisition criteria in parallel.
            **criterion_arguments: The parameters of the acquisition criterion.

        Raises:
            NotImplementedError: When the output dimension is greater than 1.
        """  # noqa: D205 D212 D415
        # Create the regressor distribution.
        if isinstance(regressor, BaseRandomProcessRegressor):
            distribution = KrigingDistribution(regressor)
            distribution.learn()
        elif isinstance(regressor, BaseRegressor):
            distribution = RegressorDistribution(regressor)
            distribution.learn()
        else:
            distribution = regressor

        if distribution.output_dimension > 1:
            msg = "ActiveLearningAlgo works only with scalar output."
            raise NotImplementedError(msg)

        # Create the acquisition problem.
        family_factory = AcquisitionCriterionFamilyFactory()
        self.__criterion_family_name = criterion_family_name
        criterion_family = family_factory.get_class(criterion_family_name)
        criterion_factory = criterion_family.ACQUISITION_CRITERION_FACTORY()
        self.__acquisition_criterion = criterion_factory.create(
            criterion_name,
            distribution,
            batch_size=batch_size,
            mc_size=mc_size,
            **criterion_arguments,
        )
        # Create the optimization space
        # that is different from the input space
        # when acquiring points in parallel.
        optimization_space = DesignSpace()
        lower_bound = tile(input_space.get_lower_bounds(), batch_size)
        upper_bound = tile(input_space.get_upper_bounds(), batch_size)
        input_space.initialize_missing_current_values()
        value = tile(input_space.get_current_value(), batch_size)
        optimization_space.add_variable(
            "x",
            size=len(lower_bound),
            lower_bound=lower_bound,
            upper_bound=upper_bound,
            value=value,
        )
        problem = self.__acquisition_problem = OptimizationProblem(optimization_space)
        problem.objective = self.__acquisition_criterion
        if not problem.objective.has_jac:
            problem.differentiation_method = (
                OptimizationProblem.DifferentiationMethod.FINITE_DIFFERENCES
            )
        if problem.objective.MAXIMIZE:
            problem.minimize_objective = False

        # Initialize acquisition algorithm.
        self.set_acquisition_algorithm(
            self.__default_algo_name, **self.__default_algo_settings
        )

        # Miscellaneous.
        self.__database = Database()
        self.__n_initial_samples = len(distribution.algo.learning_set)
        self.__distribution = distribution
        self.__input_space = input_space
        self.__batch_size = batch_size

        # Create the acquisition view.
        if distribution.algo.input_dimension == 2 and batch_size == 1:
            self.__acquisition_view = AcquisitionView(self)
        else:
            self.__acquisition_view = None

        self.__n_evaluations_history = []
        self.__qoi_history = []

    @property
    def qoi(self) -> float | None:
        """The quantity of interest (QOI).

        When there is no quantity of interest associated with the acquisition criterion,
        this attribute is `None`.
        """
        return (
            None if self.__acquisition_criterion.qoi is None else self.__qoi_history[-1]
        )

    @property
    def qoi_history(self) -> tuple[list[int], list[float]]:
        """The history of the quantity of interest (QOI) when it exists.

        The second term represents this history while the first one represents the
        history of the number of model evaluations corresponding to these QOI
        estimations.

        When there is no quantity of interest associated with the acquisition criterion,
        these lists are empty.
        """
        return self.__n_evaluations_history, self.__qoi_history

    @property
    def acquisition_criterion(self) -> BaseAcquisitionCriterion:
        """The acquisition criterion."""
        return self.__acquisition_criterion

    @property
    def acquisition_criterion_family_name(self) -> str:
        """The name of the acquisition criterion family."""
        return self.__criterion_family_name

    @property
    def regressor(self) -> BaseRegressor:
        """The regressor."""
        return self.__distribution.algo

    @property
    def regressor_distribution(self) -> BaseRegressorDistribution:
        """The regressor distribution."""
        return self.__distribution

    @property
    def n_initial_samples(self) -> int:
        """The number of initial samples."""
        return self.__n_initial_samples

    @property
    def input_space(self) -> DesignSpace:
        """The input space."""
        return self.__input_space

    @property
    def batch_size(self) -> int:
        """The number of points to be acquired in parallel."""
        return self.__batch_size

    def set_acquisition_algorithm(self, algo_name: str, **settings: Any) -> None:
        """Set sampling or optimization algorithm.

        Args:
            algo_name: The name of a DOE or optimization algorithm
                to find the learning point(s).
            **settings: The values of some algorithm settings.
        """
        factory = DOELibraryFactory()
        if not factory.is_available(algo_name):
            factory = OptimizationLibraryFactory()

        self.__acquisition_algo = factory.create(algo_name)
        self.__acquisition_algo_settings = settings

    def find_next_point(
        self,
        as_dict: bool = False,
    ) -> DataType:
        """Find the next `batch_size` learning point(s).

        Args:
            as_dict: Whether to return the input data split by input names.
                Otherwise, return a unique array.
                In both cases,
                the arrays will be shaped as ``(batch_size, input_dimension)``.

        Returns:
            The next `batch_size` learning point(s).
        """
        with LoggingContext(logging.getLogger("gemseo")):
            input_data = self.__acquisition_algo.execute(
                self.__acquisition_problem, **self.__acquisition_algo_settings
            ).x_opt
            input_data = input_data.reshape(self.__batch_size, -1)
        if as_dict:
            return self.__acquisition_problem.design_space.convert_array_to_dict(
                input_data
            )

        return input_data

    def acquire_new_points(
        self,
        discipline: Discipline,
        n_samples: int = 1,
        show: bool = False,
        file_path: str | Path = "",
    ) -> tuple[Database, OptimizationProblem]:
        """Update the machine learning algorithm by learning new samples.

        This method acquires new learning input-output samples
        and trains the machine learning algorithm
        with the resulting enriched learning set.
        The effective number of points will be the largest integer multiple
        of batch_size and less than or equal to n_samples.

        Args:
            discipline: The discipline computing the reference output data
                from the input data provided by the acquisition process.
            n_samples: The number of samples to update the machine learning algorithm.
                It should be a multiple of batch_size.
            show: Whether to display intermediate results
                Only when the input space dimension is 2.
            file_path: The file path to save the plots of the intermediate results.
                If empty, do not save them.
                Only when the input space dimension is 2.

        Returns:
            The concatenation of the optimization histories
            related to the different points
            and the last acquisition problem.

        Raises:
            ValueError: When the input space dimension is not 2.
        """
        plot = show or file_path
        if plot:
            self.__check_acquisition_view()

        self.__n_evaluations_history.append(self.__n_initial_samples)
        self.__qoi_history.append(self.__acquisition_criterion.qoi)
        total_n_samples = self.__n_initial_samples
        n_batches = int(n_samples / self.__batch_size)
        LOGGER.info("Acquiring %s points in batches of %s", n_samples, self.batch_size)
        with OneLineLogging(TQDM_LOGGER):
            for batch_id in CustomTqdmProgressBar(range(1, n_batches + 1)):
                array_input_data = self.find_next_point()
                if plot:
                    self.__acquisition_view.draw(
                        discipline=discipline,
                        new_point=array_input_data[0],
                        show=show,
                        file_path=file_path,
                    )

                for inputs, outputs in self.__acquisition_problem.database.items():
                    self.__database.store(
                        array([batch_id, *inputs.unwrap().tolist()]), outputs
                    )

                for points in range(self.__batch_size):
                    input_data = self.__input_space.convert_array_to_dict(
                        array_input_data[points, :]
                    )

                    discipline.execute(input_data)

                    extra_learning_set = IODataset()
                    distribution = self.__distribution
                    variable_names_to_n_components = distribution.algo.sizes
                    new_points = hstack(list(input_data.values()))[newaxis]
                    extra_learning_set.add_group(
                        group_name=IODataset.INPUT_GROUP,
                        data=new_points,
                        variable_names=distribution.input_names,
                        variable_names_to_n_components=variable_names_to_n_components,
                    )

                    output_names = distribution.output_names
                    output_data = discipline.get_output_data()
                    extra_learning_set.add_group(
                        group_name=IODataset.OUTPUT_GROUP,
                        data=hstack([output_data[name] for name in output_names])[
                            newaxis
                        ],
                        variable_names=output_names,
                        variable_names_to_n_components=variable_names_to_n_components,
                    )

                    augmented_learning_set = concat(
                        [distribution.algo.learning_set, extra_learning_set],
                        ignore_index=True,
                    )

                    self.__distribution.change_learning_set(augmented_learning_set)

                self.update_problem()
                self.__qoi_history.append(self.__acquisition_criterion.qoi)
                total_n_samples += self.__batch_size
                self.__n_evaluations_history.append(total_n_samples)

        return self.__database, self.__acquisition_problem

    def __check_acquisition_view(self) -> None:
        """Check that the acquisition view can be used.

        Raises:
            ValueError: When the input space dimension is not 2.
        """
        if not self.__acquisition_view:
            msg = (
                "Plotting intermediate results "
                "requires an input space dimension equal to 2."
            )
            raise ValueError(msg)

    def plot_acquisition_view(
        self,
        discipline: Discipline | None = None,
        filled: bool = True,
        n_test: int = 30,
        show: bool = True,
        file_path: str | Path = "",
    ) -> Figure:
        """Draw the points acquired through active learning.

        This visualization includes four surface plots
        representing variables of interest in function of the two inputs:

        - (top left) the surface plot of the discipline,
        - (top right) the surface plot of the acquisition criterion,
        - (bottom left) the surface plot of the regressor,
        - (bottom right) the standard deviation of the regressor.

        The $N$ data used to initialize the regressor are represented by small dots,
        the point optimizing the acquisition criterion is represented by a big dot
        (this will be the next point to learn)
        and the points added by the active learning procedure are represented
        by plus signs and labeled by their position in the learning dataset.

        Args:
            discipline: The discipline for which `n_test**2` evaluations will be done.
                If `None`,
                do not plot the discipline, i.e. the first subplot.
                In particular,
                if the discipline is costly,
                it is better to leave this argument to `None`.
            n_test: The number of points per dimension.
            filled: Whether to plot filled contours.
                Otherwise, plot contour lines.
            show: Whether to display the view.
            file_path: The file path to save the view.
                If empty, do not save it.

        Returns:
            The acquisition view.

        Raises:
            ValueError: When the input space dimension is not 2.
        """
        self.__check_acquisition_view()
        return self.__acquisition_view.draw(
            discipline=discipline,
            filled=filled,
            n_test=n_test,
            show=show,
            file_path=file_path,
        )

    def plot_qoi_history(
        self,
        show: bool = True,
        file_path: str | Path = "",
        label: str = "",
        add_markers: bool = True,
        **options: Any,
    ) -> Lines:
        """Plot the history of the quantity of interest.

        Args:
            show: Whether to display the view.
            file_path: The file path to save the view.
                If empty, do not save it.
            label: The label for the QOI.
                If empty, the name of the criterion family name.
            add_markers: Whether to add markers.
            **options: The options to create the
                [Lines][gemseo.post.dataset.lines.Lines] object.

        Returns:
            The history of the quantity of interest.
        """
        return QOIHistoryView(self).draw(
            show=show,
            file_path=file_path,
            label=label,
            add_markers=add_markers,
            **options,
        )

    def update_problem(self) -> None:
        """Update the acquisition problem."""
        self.__acquisition_problem.reset(preprocessing=False)
        self.__acquisition_criterion.update()



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Contributors:
#    AUTHORS:
#       - Francois Gallard
"""Parallel evaluation on multiple processes or threads."""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Callable

from gemseo.core.parallel_execution.callable_parallel_execution import (
    CallableParallelExecution,
)
from numpy import array
from smt.applications.ego import Evaluator

if TYPE_CHECKING:
    from gemseo.typing import RealArray


class ParallelEvaluator(Evaluator):
    """Parallel evaluation on multiple processes or threads."""

    __n_processes: int
    """The number of processes or threads to use."""

    def __init__(self, n_processes: int) -> None:
        """
        Args:
            n_processes: The number of processes or threads to use.
        """  # noqa: D205 D212
        super().__init__()
        self.__n_processes = n_processes

    def run(
        self,
        function: Callable[[RealArray], float],
        input_values: RealArray,
    ) -> RealArray:
        """Evaluate a function on a sample of input points.

        Args:
            function: The function to evaluate.
            input_values: The sample of input points.

        Returns:
            The function evaluations on the sample of points.
        """
        if (n_processes := self.__n_processes) > 1:
            executor = CallableParallelExecution((function,), n_processes=n_processes)
            output_values = executor.execute(input_values)
        else:
            output_values = [function(input_value) for input_value in input_values]

        return array([output_values], dtype="float64").T



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Mixin for confidence bounds."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from gemseo.typing import NumberArray

    from gemseo_mlearning.active_learning.distributions import BaseRegressorDistribution


class ConfidenceBound:
    """A mixin for confidence bounds."""

    __kappa: float
    """The factor associated with the standard deviation."""

    def __init__(
        self,
        regressor_distribution: BaseRegressorDistribution,
        kappa: float = 2.0,
        batch_size: int = 1,
        mc_size: int = 10_000,
    ) -> None:
        """
        Args:
            regressor_distribution: The distribution of the regressor.
            kappa: A factor associated with the standard deviation
                to increase the mean value.
            batch_size: The number of points to be acquired in parallel;
                if `1`, acquire points sequentially.
            mc_size: The sample size to estimate the acquisition criterion in parallel.
        """  # noqa: D205 D212 D415
        self.__kappa = kappa
        super().__init__(regressor_distribution, batch_size=batch_size, mc_size=mc_size)

    def _compute(self, input_value: NumberArray) -> NumberArray | float:  # noqa: D102
        mean = self._compute_mean(input_value)
        sigma = self._compute_standard_deviation(input_value)
        return (mean + self.__kappa * sigma) / self._scaling_factor



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
from __future__ import annotations

import re
from typing import TYPE_CHECKING

import pytest
from gemseo.algos.doe.scipy.scipy_doe import SciPyDOE
from gemseo.core.mdo_functions.mdo_function import MDOFunction
from gemseo.datasets.io_dataset import IODataset
from gemseo.problems.optimization.rosenbrock import Rosenbrock
from numpy import array
from numpy import atleast_2d
from numpy import concatenate
from numpy import full
from numpy import hstack
from numpy import linspace
from numpy import newaxis
from numpy.testing import assert_almost_equal
from smt.surrogate_models import GEKPLS
from smt.surrogate_models import RBF

from gemseo_mlearning.regression.smt_regressor import _NAMES_TO_CLASSES
from gemseo_mlearning.regression.smt_regressor import SMTRegressor
from gemseo_mlearning.regression.smt_regressor_settings import SMT_Regressor_Settings
from gemseo_mlearning.regression.smt_regressor_settings import SMTRegressorSettings


@pytest.mark.parametrize("name", ["RBF", "KRG"])
def test_available_models(name):
    """Check the SMT's surrogate models that can be instantiated."""
    assert name in _NAMES_TO_CLASSES


if TYPE_CHECKING:
    from gemseo.typing import RealArray


@pytest.fixture(scope="module", params=[False, True])
def io_data(request) -> tuple[RealArray, RealArray]:
    """Input and output data."""
    if request.param:
        x = linspace(-1, 1, 10)[:, newaxis]
        y = x**2
    else:
        x = linspace(-1, 1, 10).reshape((5, 2))
        y = hstack(((x[:, [0]] - x[:, [1]]) ** 2, (x[:, [0]] + x[:, [1]]) ** 2))

    return x, y


@pytest.fixture(scope="module")
def regressor_and_model(io_data) -> tuple[SMTRegressor, RBF]:
    """The SMT-based regressor and the corresponding SMT model."""
    input_data, output_data = io_data

    dataset = IODataset()
    dataset.add_input_group(input_data)
    dataset.add_output_group(output_data)

    regressor = SMTRegressor(dataset, model_class_name="RBF", parameters={"d0": 2})
    regressor.learn()

    model = RBF(d0=2)
    model.set_training_values(input_data, output_data)
    model.train()
    return regressor, model


@pytest.fixture(scope="module", params=[False, True])
def input_value(regressor_and_model, request) -> RealArray:
    """The input value."""
    input_value_ = full((regressor_and_model[0].output_dimension,), 0.25)
    if request.param:
        return input_value_[newaxis, :]

    return input_value_


def test_smt_regression_model_predict(regressor_and_model, input_value):
    """Check SMTRegressor.predict method."""
    regressor, model = regressor_and_model
    expected = model.predict_values(atleast_2d(input_value))
    if input_value.ndim == 1:
        expected = expected[0]

    assert_almost_equal(regressor.predict(input_value), expected)


def test_smt_regression_model_predict_jacobian(regressor_and_model, input_value):
    """Check SMTRegressor.predict_jacobian method."""
    regressor, model = regressor_and_model
    expected = model.predict_derivatives(atleast_2d(input_value), 0)[..., newaxis]
    if regressor.output_dimension == 2:
        expected = concatenate(
            (
                expected,
                model.predict_derivatives(atleast_2d(input_value), 1)[..., newaxis],
            ),
            axis=2,
        )

    if input_value.ndim == 1:
        expected = expected[0]

    assert_almost_equal(regressor.predict_jacobian(input_value), expected)


def test_gradient_enhanced_smt_regressor():
    """Check that gradient-enhanced SMT surrogate models use the gradient samples."""

    def compute_sum_jacobian(input_value: RealArray) -> RealArray:
        """Compute the Jacobian of the sum function at a given point.

        Args:
            input_value: The input value.

        Returns:
            The Jacobian.
        """
        return array([[1.0, 1.0]])

    problem = Rosenbrock()
    problem.add_observable(MDOFunction(sum, "sum", jac=compute_sum_jacobian))
    SciPyDOE("LHS").execute(problem, eval_jac=True, n_samples=30)
    dataset = problem.to_dataset(opt_naming=False, export_gradients=True)

    regressor = SMTRegressor(dataset, model_class_name="GEKPLS")

    # Check that the training does not raise an exception
    regressor.learn()

    # Check the default values
    algo = regressor.algo
    assert isinstance(algo, GEKPLS)
    assert algo.options["n_comp"] == 2

    # Check that custom values can be passed
    regressor = SMTRegressor(
        dataset, model_class_name="GEKPLS", parameters={"n_comp": 5}
    )
    assert regressor.algo.options["n_comp"] == 5

    # Check that passing a gradient-free IODataset raises a ValueError
    x = linspace(-1, 1, 10)[:, newaxis]
    y = x**2
    dataset = IODataset()
    dataset.add_input_group(x)
    dataset.add_output_group(y)

    regressor = SMTRegressor(dataset, model_class_name="GEKPLS")
    with pytest.raises(
        ValueError,
        match=re.escape(
            "GEKPLS did not found gradient samples in the training dataset."
        ),
    ):
        regressor.learn()


def test_alias():
    """Verify that SMTRegressorSettings is an alias of SMT_Regressor_Settings."""
    assert SMTRegressorSettings == SMT_Regressor_Settings



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""A library for surrogate-based optimization."""

from __future__ import annotations

import sys
from collections.abc import Mapping
from dataclasses import dataclass
from typing import TYPE_CHECKING
from typing import Any
from typing import ClassVar
from typing import Union

from gemseo.algos.base_driver_library import DriverSettingType
from gemseo.algos.doe.factory import DOELibraryFactory
from gemseo.algos.opt.base_optimization_library import BaseOptimizationLibrary
from gemseo.algos.opt.base_optimization_library import OptimizationAlgorithmDescription
from gemseo.mlearning.core.algos.ml_algo import MLAlgoSettingsType
from gemseo.mlearning.regression.algos.base_regressor import BaseRegressor

from gemseo_mlearning.algos.opt.core.surrogate_based_optimizer import (
    SurrogateBasedOptimizer,
)
from gemseo_mlearning.algos.opt.sbo_settings import SBO_Settings

if TYPE_CHECKING:
    from gemseo.algos.base_problem import BaseProblem


SBOSettingType = Union[
    int,
    float,
    str,
    Mapping[str, DriverSettingType],
    Mapping[str, MLAlgoSettingsType],
    Mapping[str, Any],
]


@dataclass
class SurrogateBasedAlgorithmDescription(OptimizationAlgorithmDescription):
    """The description of a surrogate-based optimization algorithm."""

    library_name: str = "gemseo-mlearning"


class SurrogateBasedOptimization(BaseOptimizationLibrary):
    """A wrapper for surrogate-based optimization."""

    LIBRARY_NAME = SurrogateBasedAlgorithmDescription.library_name

    _NORMALIZE_DS: ClassVar[bool] = False
    ALGORITHM_INFOS: ClassVar[dict[str, Any]] = {
        "SBO": SurrogateBasedAlgorithmDescription(
            algorithm_name="SBO",
            description="GEMSEO in-house surrogate-based optimizer.",
            handle_equality_constraints=False,
            handle_inequality_constraints=False,
            handle_integer_variables=True,  # provided acquisition handles integers
            internal_algorithm_name="SBO",
            Settings=SBO_Settings,
        )
    }

    def _run(
        self, problem: BaseProblem, **settings: SBOSettingType
    ) -> tuple[str, None]:
        """
        Raises:
            ValueError: When the maximum number of iterations
                is less than or equal to the initial DOE size.
        """  # noqa: D205 D212 D415
        doe_settings = settings["doe_settings"]
        doe_size = settings["doe_size"]
        doe_algorithm = settings["doe_algorithm"]
        regression_algorithm = settings["regression_algorithm"]
        if not isinstance(regression_algorithm, BaseRegressor):
            # The number of evaluations is equal to
            #     1 for the initial evaluation in OptimizationLibrary._pre_run
            #   + N for the N-length DOE
            #   + n_iter - 1 - N
            # So, n_iter - 1 - N >= 0 implies that n_iter >= 1+N
            doe_algo = DOELibraryFactory().create(doe_algorithm)
            initial_doe_size = len(
                doe_algo.compute_doe(
                    problem.design_space, n_samples=doe_size, **doe_settings
                )
            )
            max_iter = settings["max_iter"]
            if max_iter < 1 + initial_doe_size:
                msg = (
                    f"max_iter ({max_iter}) must be "
                    f"strictly greater than the initial DOE size ({initial_doe_size})."
                )
                raise ValueError(msg)

        optimizer = SurrogateBasedOptimizer(
            problem,
            settings["acquisition_algorithm"],
            doe_size=doe_size,
            doe_algorithm=doe_algorithm,
            doe_settings=doe_settings,
            regression_algorithm=regression_algorithm,
            regression_settings=settings["regression_settings"],
            regression_file_path=settings["regression_file_path"],
            **settings["acquisition_settings"],
        )
        return optimizer.execute(sys.maxsize), None



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""A problem connecting the Rosenbrock function with its uncertain space."""

from __future__ import annotations

from gemseo.algos.optimization_problem import OptimizationProblem

from gemseo_mlearning.problems.rosenbrock.rosenbrock_function import RosenbrockFunction
from gemseo_mlearning.problems.rosenbrock.rosenbrock_space import RosenbrockSpace


class RosenbrockProblem(OptimizationProblem):
    """A problem connecting the Rosenbrock function with its uncertain space."""

    def __init__(self, use_uncertain_space: bool = True) -> None:
        """
        Args:
            use_uncertain_space: Whether to consider the input space
                as an uncertain space.
        """  # noqa: D205 D212
        input_space = RosenbrockSpace()
        if not use_uncertain_space:
            input_space = input_space.to_design_space()

        super().__init__(input_space)
        self.objective = RosenbrockFunction()



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Kriging-like regressor distribution."""

from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING

from gemseo.mlearning.data_formatters.regression_data_formatters import (
    RegressionDataFormatters,
)
from scipy.stats import norm

from gemseo_mlearning.active_learning.distributions.base_regressor_distribution import (
    BaseRegressorDistribution,
)

if TYPE_CHECKING:
    from gemseo.mlearning.core.algos.ml_algo import DataType
    from gemseo.mlearning.regression.algos.base_random_process_regressor import (
        BaseRandomProcessRegressor,
    )
    from gemseo.typing import NumberArray


class KrigingDistribution(BaseRegressorDistribution):
    """Kriging-like regressor distribution."""

    algo: BaseRandomProcessRegressor

    def __init__(  # noqa: D107
        self, algo: BaseRandomProcessRegressor
    ) -> None:
        super().__init__(algo)

    def compute_confidence_interval(  # noqa: D102
        self,
        input_data: DataType,
        level: float = 0.95,
    ) -> (
        tuple[dict[str, NumberArray], dict[str, NumberArray]]
        | tuple[NumberArray, NumberArray]
    ):
        mean = self.compute_mean(input_data)
        std = self.compute_standard_deviation(input_data)
        quantile = norm.ppf(level)
        if isinstance(mean, Mapping):
            lower = {name: mean[name] - quantile * std[name] for name in mean}
            upper = {name: mean[name] + quantile * std[name] for name in mean}
        else:
            lower = mean - quantile * std
            upper = mean + quantile * std
        return lower, upper

    @RegressionDataFormatters.format_dict
    @RegressionDataFormatters.format_samples()
    def compute_mean(  # noqa: D102
        self,
        input_data: DataType,
    ) -> DataType:
        return self.algo.predict(input_data)

    @RegressionDataFormatters.format_dict
    @RegressionDataFormatters.format_samples()
    def compute_variance(  # noqa: D102
        self,
        input_data: DataType,
    ) -> DataType:
        return self.compute_standard_deviation(input_data) ** 2

    @RegressionDataFormatters.format_dict
    @RegressionDataFormatters.format_samples()
    def compute_standard_deviation(  # noqa: D102
        self,
        input_data: DataType,
    ) -> DataType:
        return self.algo.predict_std(input_data)

    def compute_samples(  # noqa: D102
        self,
        input_data: NumberArray,
        n_samples: int,
    ) -> NumberArray:
        return self.algo.compute_samples(input_data, n_samples)



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
from __future__ import annotations

import pytest
from gemseo.datasets.io_dataset import IODataset
from gemseo.mlearning.regression.algos.linreg import LinearRegressor
from gemseo.mlearning.regression.algos.rbf import RBFRegressor
from gemseo.utils.testing.helpers import concretize_classes
from numpy import array
from numpy import linspace
from numpy import newaxis

from gemseo_mlearning.active_learning.distributions.base_regressor_distribution import (
    BaseRegressorDistribution,
)
from gemseo_mlearning.active_learning.distributions.regressor_distribution import (
    RegressorDistribution,
)


@pytest.fixture(scope="module")
def distribution(dataset) -> BaseRegressorDistribution:
    """The distribution of a linear regression model built from the learning dataset."""
    with concretize_classes(BaseRegressorDistribution):
        return BaseRegressorDistribution(LinearRegressor(dataset))


@pytest.fixture(scope="module")
def distribution_with_variance(dataset) -> BaseRegressorDistribution:
    """The distribution of a linear regression model built from the learning dataset.

    This distribution has a mock implementation of the method compute_variance.
    """
    with concretize_classes(BaseRegressorDistribution):
        distribution = BaseRegressorDistribution(LinearRegressor(dataset))

    distribution.compute_variance = lambda input_data: input_data
    return distribution


@pytest.fixture(scope="module")
def distribution_with_transformers() -> RegressorDistribution:
    """The distribution of an algorithm using variable transformation."""
    dataset = IODataset()
    x = linspace(-1, 1, 10)[:, None]
    dataset.add_variable("x", x, group_name=dataset.INPUT_GROUP)
    dataset.add_variable("y", x**2, group_name=dataset.OUTPUT_GROUP)

    algo = RBFRegressor(dataset, transformer=RBFRegressor.DEFAULT_TRANSFORMER)

    distribution = RegressorDistribution(algo)
    distribution.learn()
    return distribution


def test_init(distribution):
    """Check the attributes 'algo' and '_samples' after initialization.

    Samples must be None (the value is defined by the 'learn' method) and the algorithm
    must be equal to the provided one.
    """
    assert isinstance(distribution.algo, LinearRegressor)
    assert distribution._samples == []


def test_learning_set(distribution):
    """Check the property ``learning_set``."""
    assert isinstance(distribution.learning_set, IODataset)


def test_inputs_names(distribution):
    """Check the property 'input_names'.

    Must be equal to the names of the inputs of the regression algorithm.
    """
    assert distribution.input_names == distribution.algo.input_names


def test_outputs_names(distribution):
    """Check the property 'output_names'.

    Should be equal to the names of the outputs of the regression algorithm.
    """
    assert distribution.output_names == distribution.algo.output_names


def test_output_dimension(distribution):
    """Check the property 'output_dimension'.

    Must be equal to the output dimension of the regression algorithm.
    """
    assert distribution.output_dimension == distribution.algo.output_dimension


@pytest.mark.parametrize(
    ("samples", "expected_samples", "prediction"),
    [(None, range(3), 2.0 / 3), ([0, 2], [0, 2], 1.0)],
)
def test_learn(distribution, samples, expected_samples, prediction):
    """Check the learning stage.

    The algorithm must be trained, the samples must be equal to the provided ones (or to
    all samples if None is provided) and the prediction must be right Should be equal to
    the output dimension of the regression algorithm.
    """
    distribution.learn(samples)
    assert distribution.algo.is_trained
    assert distribution._samples == expected_samples
    assert distribution.algo.predict(array([0.0]))[0] == pytest.approx(prediction, 0.01)


@pytest.mark.parametrize(
    ("input_data", "output_data"),
    [
        (array([0.0]), array([1.0])),
        (array([[0.0]]), array([[1.0]])),
        ({"x": array([0.0])}, {"y": array([1.0])}),
        ({"x": array([[0.0]])}, {"y": array([[1.0]])}),
    ],
)
def test_predict(distribution, input_data, output_data):
    """Check the prediction stage."""
    distribution.learn([0, 2])
    assert distribution.predict(input_data) == output_data


@pytest.mark.parametrize(
    ("input_data", "output_data"),
    [
        (array([4.0]), array([2.0])),
        ({"x": array([4.0])}, {"y": array([2.0])}),
        (array([[4.0]]), array([[2.0]])),
        ({"x": array([[4.0]])}, {"y": array([[2.0]])}),
    ],
)
def test_standard_deviation(distribution_with_variance, input_data, output_data):
    """Check that the standard deviation is the square root of the variance.

    This test method uses a mock distribution whose method compute_variance is the
    identity function. Thus, the compute_standard_deviation method should return the
    square root of its input.
    """
    assert (
        distribution_with_variance.compute_standard_deviation(input_data) == output_data
    )


def test_change_learning_set(dataset):
    """Check that changing the learning set updates the algorithm."""
    with concretize_classes(BaseRegressorDistribution):
        distribution = BaseRegressorDistribution(LinearRegressor(dataset))

    new_dataset = IODataset()
    new_dataset.add_variable(
        "x", array([0.0, 1.0])[:, None], group_name=new_dataset.INPUT_GROUP
    )
    new_dataset.add_variable(
        "y",
        array([1.0, 1.0])[:, None],
        group_name=new_dataset.OUTPUT_GROUP,
    )
    distribution.change_learning_set(new_dataset)
    assert len(distribution.learning_set) == 2
    assert distribution.predict(array([0.5])) == array([1.0])


@pytest.mark.parametrize("as_dict", [False, True])
@pytest.mark.parametrize("is_2d", [False, True])
@pytest.mark.parametrize(
    "statistic",
    ["variance", "mean", "standard_deviation"],
)
def test_decorator_for_statistics(
    distribution_with_transformers, as_dict, is_2d, statistic
):
    """Check the use of decorator for the methods compute_{statistic}."""
    input_data = array([0.0])
    if is_2d:
        input_data = input_data[newaxis, :]

    if as_dict:
        input_data = {"x": input_data}

    compute_statistic = getattr(distribution_with_transformers, f"compute_{statistic}")
    output_data = compute_statistic(input_data)

    assert isinstance(output_data, dict) == as_dict

    if as_dict:
        output_data = output_data["y"]

    assert output_data.ndim == 1 + is_2d



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
from __future__ import annotations

import pytest
from gemseo.datasets.io_dataset import IODataset
from gemseo.mlearning.regression.algos.gpr import GaussianProcessRegressor
from gemseo.mlearning.regression.algos.linreg import LinearRegressor
from numpy import array


@pytest.fixture(scope="module")
def dataset() -> IODataset:
    """A learning dataset with three samples."""
    dataset = IODataset()
    dataset.add_variable(
        "x", array([0.0, 0.5, 1.0])[:, None], group_name=dataset.INPUT_GROUP
    )
    dataset.add_variable(
        "y",
        array([1.0, 0.0, 1.0])[:, None],
        group_name=dataset.OUTPUT_GROUP,
    )
    return dataset


@pytest.fixture(scope="module")
def linear_algo(dataset) -> LinearRegressor:
    """A linear regression model used in different tests."""
    model = LinearRegressor(dataset)
    model.learn()
    return model


@pytest.fixture(scope="module")
def kriging_algo(dataset) -> GaussianProcessRegressor:
    """A Kriging regression model used in different tests."""
    return GaussianProcessRegressor(dataset)



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
from __future__ import annotations

import re

import pytest
from gemseo.mlearning.regression.algos.gpr import GaussianProcessRegressor
from gemseo.utils.testing.helpers import image_comparison
from matplotlib.figure import Figure
from numpy import array

from gemseo_mlearning.active_learning.active_learning_algo import ActiveLearningAlgo
from gemseo_mlearning.active_learning.visualization.acquisition_view import (
    AcquisitionView,
)


@pytest.fixture(scope="module")
def acquisition_view(active_learning_algo) -> AcquisitionView:
    """The acquisition view."""
    return AcquisitionView(active_learning_algo)


def test_raise_parallel(discipline, learning_dataset, input_space):
    """Check that an error is raised when plotting in batch mode."""
    regressor = GaussianProcessRegressor(learning_dataset)
    algo = ActiveLearningAlgo("Minimum", input_space, regressor, batch_size=3)

    with pytest.raises(
        NotImplementedError,
        match=re.escape(
            "AcquisitionView does not support active learning"
            " algorithm using parallel acquisition."
        ),
    ):
        AcquisitionView(algo)


def test_return_type(acquisition_view):
    """Check the return type of AcquisitionView."""
    assert isinstance(acquisition_view.draw(show=False), Figure)


@image_comparison(["default"], tol=0.9)
def test_default(acquisition_view):
    """Check AcquisitionView with default settings."""
    acquisition_view.draw(show=False)


@image_comparison(["custom"], tol=0.9)
def test_custom(acquisition_view, discipline):
    """Check AcquisitionView with custom settings."""
    acquisition_view.draw(
        show=False,
        discipline=discipline,
        new_point=array([0.0, 0.0]),
        filled=False,
        n_test=5,
    )


@image_comparison(["new_point"], tol=0.9)
def test_new_point(acquisition_view):
    """Check AcquisitionView with a new point."""
    acquisition_view.draw(show=False, new_point=array([0.0, 0.0]))


@image_comparison(["filled_false"], tol=0.9)
def test_filled(acquisition_view):
    """Check AcquisitionView without filled contours."""
    acquisition_view.draw(show=False, filled=False)


@image_comparison(["n_test"], tol=0.9)
def test_n_test(acquisition_view):
    """Check AcquisitionView with a lower number of points."""
    acquisition_view.draw(show=False, n_test=5)


@image_comparison(["discipline"], tol=0.9)
def test_discipline(acquisition_view, discipline):
    """Check AcquisitionView with a discipline."""
    acquisition_view.draw(show=False, discipline=discipline)



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
from __future__ import annotations

import re
from operator import eq
from pathlib import Path
from unittest import mock

import pytest
from gemseo.algos.design_space import DesignSpace
from gemseo.algos.optimization_problem import OptimizationProblem
from gemseo.algos.parameter_space import ParameterSpace
from gemseo.datasets.io_dataset import IODataset
from gemseo.disciplines.analytic import AnalyticDiscipline
from gemseo.mlearning.regression.algos.gpr import GaussianProcessRegressor
from gemseo.mlearning.regression.algos.linreg import LinearRegressor
from numpy import array
from numpy import ndarray
from numpy.testing import assert_almost_equal

from gemseo_mlearning.active_learning.acquisition_criteria.minimum.base_minimum import (
    BaseMinimum,
)
from gemseo_mlearning.active_learning.active_learning_algo import ActiveLearningAlgo
from gemseo_mlearning.active_learning.distributions import KrigingDistribution
from gemseo_mlearning.active_learning.distributions.regressor_distribution import (
    RegressorDistribution,
)
from gemseo_mlearning.active_learning.visualization.acquisition_view import (
    AcquisitionView,
)
from gemseo_mlearning.active_learning.visualization.qoi_history_view import (
    QOIHistoryView,
)


@pytest.fixture(scope="module")
def dataset() -> IODataset:
    """A learning dataset."""
    dataset = IODataset()
    dataset.add_variable(
        "x", array([0.0, 1.0])[:, None], group_name=dataset.INPUT_GROUP
    )
    dataset.add_variable(
        "y",
        array([1.0, 2.0])[:, None],
        group_name=dataset.OUTPUT_GROUP,
    )
    return dataset


@pytest.fixture(scope="module")
def kriging_distribution(dataset) -> KrigingDistribution:
    """A Kriging distribution."""
    distribution = KrigingDistribution(GaussianProcessRegressor(dataset, alpha=0.0))
    distribution.learn()
    return distribution


@pytest.fixture
def algo_distribution_for_update(dataset) -> RegressorDistribution:
    """The distribution of a linear regression model to be updated."""
    distribution = RegressorDistribution(LinearRegressor(dataset), size=3)
    distribution.learn()
    return distribution


@pytest.fixture(scope="module")
def algo_distribution(dataset) -> RegressorDistribution:
    """The distribution of a linear regression model."""
    distribution = RegressorDistribution(LinearRegressor(dataset), size=3)
    distribution.learn()
    return distribution


@pytest.fixture(scope="module")
def input_space() -> DesignSpace:
    """The input space used to acquire new points."""
    space = DesignSpace()
    space.add_variable("x", lower_bound=0.0, upper_bound=1.0)
    return space


def test_init(algo_distribution, input_space):
    """Check the initialization of the ActiveLearningAlgo for sequential active
    learning.
    """
    algo = ActiveLearningAlgo("Minimum", input_space, algo_distribution)
    assert algo.n_initial_samples == 2
    assert algo.regressor_distribution == algo_distribution
    assert algo.regressor == algo_distribution.algo
    assert algo.input_space == input_space
    assert algo._ActiveLearningAlgo__batch_size == 1
    assert algo.acquisition_criterion._mc_size == 10000
    assert algo.acquisition_criterion.name == "EI"
    assert algo._ActiveLearningAlgo__acquisition_problem.objective.name == "-EI"
    assert (
        algo._ActiveLearningAlgo__acquisition_algo.algo_name
        == algo._ActiveLearningAlgo__default_algo_name
    )
    assert (
        algo._ActiveLearningAlgo__acquisition_algo_settings
        == algo._ActiveLearningAlgo__default_algo_settings
    )
    assert (
        algo._ActiveLearningAlgo__acquisition_problem.design_space.variable_names
        == ["x"]
    )
    assert algo._ActiveLearningAlgo__acquisition_problem.design_space == input_space
    assert algo._ActiveLearningAlgo__distribution == algo_distribution
    assert algo._ActiveLearningAlgo__acquisition_algo.algo_name == "MultiStart"


def test_init_parallel(kriging_distribution, input_space):
    """Check the initialization of the ActiveLearningAlgo for parallel active
    learning.
    """
    algo = ActiveLearningAlgo(
        "Minimum", input_space, kriging_distribution, batch_size=2
    )
    assert algo.input_space.variable_names == input_space.variable_names
    assert len(
        algo._ActiveLearningAlgo__acquisition_problem.design_space.get_lower_bounds()
    ) / 2 == len(input_space.variable_names)
    assert algo._ActiveLearningAlgo__batch_size == 2


def test_init_with_bad_output_dimension(input_space):
    """Check ActiveLearningAlgo initialization raising errors.

    The initialization should raise a NotImplementedError when the output dimension of
    the algorithm is greater than 1.
    """
    dataset = IODataset()
    dataset.add_variable("x", array([[0.0], [1.0]]), group_name=dataset.INPUT_GROUP)
    dataset.add_variable(
        "y",
        array([[1.0, 1.0], [2.0, 2.0]]),
        group_name=dataset.OUTPUT_GROUP,
    )
    algo = LinearRegressor(dataset)
    algo.learn()
    algo_distribution = RegressorDistribution(algo, size=3)
    with pytest.raises(
        NotImplementedError,
        match=re.escape("ActiveLearningAlgo works only with scalar output."),
    ):
        ActiveLearningAlgo("Minimum", input_space, algo_distribution)


@pytest.mark.parametrize(
    ("criterion_family_name", "criterion_name"),
    [
        ("Minimum", "Mean"),
        ("Maximum", "Mean"),
        ("Maximum", "UCB"),
        ("Minimum", "LCB"),
        ("Exploration", "Variance"),
        ("Exploration", "Distance"),
        ("Exploration", "StandardDeviation"),
    ],
)
def test_with_bad_parallelization(
    kriging_distribution, input_space, criterion_family_name, criterion_name
):
    """Check parallelization raising errors.

    A NotImplementedError is raised when parallelization is considered for criteria not
    compatible with this method.
    """
    active_learning = ActiveLearningAlgo(
        criterion_family_name=criterion_family_name,
        criterion_name=criterion_name,
        input_space=input_space,
        regressor=kriging_distribution,
        batch_size=2,
    )

    with pytest.raises(
        NotImplementedError,
        match=re.escape(
            "Parallelization is not defined for this acquisition criterion."
        ),
    ):
        active_learning.find_next_point()


@pytest.mark.parametrize(
    ("algo_name", "setting_name", "setting_value"),
    [
        ("PYDOE_FULLFACT", "n_samples", 3),
        ("SLSQP", "max_iter", 3),
        ("PYDOE_FULLFACT", "n_samples", None),
        ("SLSQP", "max_iter", None),
    ],
)
def test_set_acquisition_algorithm(
    algo_distribution, input_space, algo_name, setting_name, setting_value
):
    """Check the setting of the acquisition algorithm used by the ActiveLearningAlgo."""
    algo = ActiveLearningAlgo("Minimum", input_space, algo_distribution)
    kwargs = {"algo_name": algo_name}
    settings = {}
    if setting_value is not None:
        settings[setting_name] = setting_value
        kwargs.update(settings)

    algo.set_acquisition_algorithm(**kwargs)
    assert algo._ActiveLearningAlgo__acquisition_algo.algo_name == algo_name
    assert algo._ActiveLearningAlgo__acquisition_algo_settings == settings


@pytest.mark.parametrize("as_dict", [False, True])
@pytest.mark.parametrize("batch_size", [1, 2])
def test_compute_parallel(kriging_distribution, input_space, as_dict, batch_size):
    """Check the acquisition of a new point by the ActiveLearningAlgo."""
    algo = ActiveLearningAlgo(
        criterion_family_name="Minimum",
        input_space=input_space,
        regressor=kriging_distribution,
        batch_size=batch_size,
    )
    algo.set_acquisition_algorithm("PYDOE_FULLFACT", n_samples=3)
    x_opt = algo.find_next_point(as_dict=as_dict)
    x_opt = x_opt if isinstance(x_opt, ndarray) else x_opt["x"]
    assert x_opt.shape == (batch_size, len(input_space.variable_names))


@pytest.mark.parametrize("criterion_family_name", ["Exploration", "Maximum"])
def test_update_algo(algo_distribution_for_update, input_space, criterion_family_name):
    """Check the update of the machine learning algorithm."""
    distribution = algo_distribution_for_update
    initial_size = len(distribution.learning_set)
    algo = ActiveLearningAlgo(criterion_family_name, input_space, distribution)
    algo.set_acquisition_algorithm("PYDOE_FULLFACT", n_samples=3)
    # z values will not be used
    # because the initial surrogate does not have z as an output.
    # Nevertheless,
    # we add this output to ensure that ActiveLearningAlgo filters the outputs.
    algo.acquire_new_points(AnalyticDiscipline({"y": "1+x", "z": "1+x"}))
    if criterion_family_name == "Exploration":
        assert algo.qoi is None
    else:
        assert_almost_equal(algo.qoi, array([2.0]))
    assert len(distribution.learning_set) == initial_size + 1


@pytest.mark.parametrize(
    ("criterion", "minimize"),
    [("Quantile", True), ("Minimum", False)],
)
def test_build_opt_problem_maximize(
    algo_distribution, input_space, criterion, minimize
):
    """Check that the optimization problem handles both cost & performance criteria."""
    kwargs = {}
    if criterion == "Quantile":
        kwargs["level"] = 0.1
        uncertain_space = ParameterSpace()
        uncertain_space.add_random_variable("x", "OTNormalDistribution")
        kwargs["uncertain_space"] = uncertain_space
    algo = ActiveLearningAlgo(criterion, input_space, algo_distribution, **kwargs)
    assert algo._ActiveLearningAlgo__acquisition_problem.minimize_objective == minimize


@pytest.mark.parametrize(
    ("criterion", "kwargs", "use_finite_differences"),
    [("Maximum", {}, True), ("Minimum", {}, True)],
)
def test_build_opt_problem_jacobian(
    algo_distribution, input_space, criterion, use_finite_differences, kwargs
):
    """Check that the optimization problem can use approximated or analytic Jacobian."""
    algo = ActiveLearningAlgo(criterion, input_space, algo_distribution, **kwargs)
    assert (
        eq(
            algo._ActiveLearningAlgo__acquisition_problem.differentiation_method,
            OptimizationProblem.DifferentiationMethod.FINITE_DIFFERENCES,
        )
        == use_finite_differences
    )


@pytest.mark.parametrize(
    ("has_jac", "differentiation_method"),
    [
        (False, OptimizationProblem.DifferentiationMethod.FINITE_DIFFERENCES),
        (True, OptimizationProblem.DifferentiationMethod.USER_GRAD),
    ],
)
def test_analytical_jacobian(
    algo_distribution, input_space, has_jac, differentiation_method
):
    """Check that the OptimizationProblem uses the analytic Jacobian when available."""
    with mock.patch.object(BaseMinimum, "has_jac", has_jac):
        algo = ActiveLearningAlgo("Minimum", input_space, algo_distribution)

    assert (
        algo._ActiveLearningAlgo__acquisition_problem.differentiation_method
        == differentiation_method
    )


@pytest.mark.parametrize(
    ("regressor_class", "regressor_distribution_class"),
    [
        (LinearRegressor, RegressorDistribution),
        (GaussianProcessRegressor, KrigingDistribution),
    ],
)
def test_regressor_at_instantiation(
    input_space, dataset, regressor_class, regressor_distribution_class
):
    """Check that ActiveLearningAlgo can be instantiated from a regressor."""
    regressor = regressor_class(dataset)
    algo = ActiveLearningAlgo("Minimum", input_space, regressor)
    distribution = algo._ActiveLearningAlgo__distribution
    assert isinstance(distribution, regressor_distribution_class)
    assert distribution.algo == regressor


@pytest.fixture
def algo_for_plotting(algo_distribution_for_update) -> ActiveLearningAlgo:
    """An active learning algorithm to test plotting methods."""
    dataset = IODataset()
    dataset.add_variable(
        "x1", array([[0.0], [1.0], [2.0]]), group_name=dataset.INPUT_GROUP
    )
    dataset.add_variable(
        "x2", array([[1.0], [2.0], [3.0]]), group_name=dataset.INPUT_GROUP
    )
    dataset.add_variable(
        "y",
        array([[1.0], [2.0], [3.0]]),
        group_name=dataset.OUTPUT_GROUP,
    )
    regressor = GaussianProcessRegressor(dataset)
    space = DesignSpace()
    space.add_variable("x1", lower_bound=0.0, upper_bound=1.0, value=0.5)
    space.add_variable("x2", lower_bound=0.0, upper_bound=1.0, value=0.5)
    return ActiveLearningAlgo("Minimum", space, regressor)


def test_online_plot(algo_for_plotting, tmp_wd):
    """Check that a plot is generated when acquiring new points."""
    algo_for_plotting.acquire_new_points(
        AnalyticDiscipline({"y": "1+x1+x2"}), file_path=Path("foo.png")
    )
    with mock.patch.object(AcquisitionView, "draw") as draw:
        algo_for_plotting.plot_acquisition_view()

    assert draw.call_args.kwargs == {
        "discipline": None,
        "file_path": "",
        "filled": True,
        "n_test": 30,
        "show": True,
    }


@pytest.mark.parametrize("input_dimension", [1, 3])
def test_online_plot_error(algo_distribution_for_update, input_dimension):
    """Check that an error is raised when plotting in dimension != 2."""
    dataset = IODataset()
    space = DesignSpace()
    for i in range(input_dimension):
        dataset.add_variable(
            f"x{i}", array([[0.0], [1.0], [2.0]]), group_name=dataset.INPUT_GROUP
        )
        space.add_variable(f"x{i}", lower_bound=0.0, upper_bound=1.0, value=0.5)

    dataset.add_variable(
        "y",
        array([[1.0], [2.0], [3.0]]),
        group_name=dataset.OUTPUT_GROUP,
    )

    algo = ActiveLearningAlgo("Minimum", space, GaussianProcessRegressor(dataset))
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Plotting intermediate results "
            "requires an input space dimension equal to 2."
        ),
    ):
        algo.acquire_new_points(AnalyticDiscipline({"y": "1+x1"}), show=True)


def test_plot_qoi_history(input_space, algo_distribution_for_update):
    """Check that plot_qoi_history works correctly."""
    algo = ActiveLearningAlgo("Minimum", input_space, algo_distribution_for_update)
    algo.acquire_new_points(AnalyticDiscipline({"y": "1+x"}))
    with mock.patch.object(QOIHistoryView, "draw") as draw:
        algo.plot_qoi_history()

    assert draw.call_args.kwargs == {
        "add_markers": True,
        "file_path": "",
        "label": "",
        "show": True,
    }

    with mock.patch.object(QOIHistoryView, "draw") as draw:
        algo.plot_qoi_history(show=1, file_path=2, label=3, add_markers=4)

    assert draw.call_args.kwargs == {
        "add_markers": 4,
        "file_path": 2,
        "label": 3,
        "show": 1,
    }


def test_plot_acquisition_view(algo_for_plotting):
    """Check that plot_acquisition_view works correctly."""
    algo_for_plotting.acquire_new_points(AnalyticDiscipline({"y": "1+x1+x2"}))
    with mock.patch.object(AcquisitionView, "draw") as draw:
        algo_for_plotting.plot_acquisition_view()

    assert draw.call_args.kwargs == {
        "discipline": None,
        "file_path": "",
        "filled": True,
        "n_test": 30,
        "show": True,
    }

    with mock.patch.object(AcquisitionView, "draw") as draw:
        algo_for_plotting.plot_acquisition_view(
            discipline=1, filled=2, n_test=3, show=4, file_path=5
        )

    assert draw.call_args.kwargs == {
        "discipline": 1,
        "file_path": 5,
        "filled": 2,
        "n_test": 3,
        "show": 4,
    }



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
from __future__ import annotations

import re

import pytest
from numpy import array
from numpy.testing import assert_almost_equal

from gemseo_mlearning.active_learning.acquisition_criteria.maximum.ei import EI
from gemseo_mlearning.active_learning.acquisition_criteria.maximum.mean import Mean
from gemseo_mlearning.active_learning.acquisition_criteria.maximum.ucb import UCB


@pytest.mark.parametrize(
    ("cls", "kwargs", "input_value", "expected"),
    [
        (EI, {}, array([0.0]), array([0.0])),
        (UCB, {}, array([0.0]), array([1.0])),
        (UCB, {"kappa": 3.0}, array([0.0]), array([1.0])),
        (Mean, {}, array([0.0]), array([1.0])),
        (EI, {}, array([[0.0]] * 2), array([[0.0]] * 2)),
        (UCB, {}, array([[0.0]] * 2), array([[1.0]] * 2)),
        (UCB, {"kappa": 3.0}, array([[0.0]] * 2), array([[1.0]] * 2)),
        (Mean, {}, array([[0.0]] * 2), array([[1.0]] * 2)),
    ],
)
def test_maximum_kriging_regressor(
    kriging_distribution, cls, kwargs, input_value, expected
):
    """Check the criteria deriving from BaseMaximum with a Kriging distribution."""
    criterion = cls(kriging_distribution, **kwargs)
    assert_almost_equal(criterion.func(input_value), expected)
    assert criterion._mc_size == 10000
    assert criterion._batch_size == 1


@pytest.mark.parametrize(
    ("cls", "kwargs", "input_value", "expected"),
    [
        (EI, {}, array([0.123]), array([0.0])),
        (UCB, {}, array([0.123]), array([1.3609677])),
        (UCB, {"kappa": 3.0}, array([0.123]), array([1.8569516])),
        (Mean, {}, array([0.123]), array([0.369])),
        (EI, {}, array([[0.123]] * 2), array([[0.0]] * 2)),
        (UCB, {}, array([[0.123]] * 2), array([[1.3609677]] * 2)),
        (UCB, {"kappa": 3.0}, array([[0.123]] * 2), array([[1.8569516]] * 2)),
        (Mean, {}, array([[0.123]] * 2), array([[0.369]] * 2)),
    ],
)
def test_maximum_regressor(algo_distribution, kwargs, cls, input_value, expected):
    """Check the criteria deriving from BaseMaximum with a RegressorDistribution."""
    criterion = cls(algo_distribution, **kwargs)
    assert_almost_equal(criterion.func(input_value), expected)
    assert criterion._mc_size == 10000
    assert criterion._batch_size == 1


@pytest.mark.parametrize(
    ("kwargs", "input_value", "expected"),
    [
        ({}, array([[1.0]] * 2), 0),
        ({}, array([[3]] * 2), 1.533606426240575),
        ({"mc_size": 50000}, array([[3]] * 2), 1.8365297109444447),
    ],
)
def test_maximum_parallel_kriging_regressor(
    kriging_distribution, kwargs, input_value, expected
):
    """Check the parallelized criteria deriving from BaseMaximum."""
    criterion = EI(kriging_distribution, batch_size=2, **kwargs)
    assert_almost_equal(criterion.func(input_value), expected)
    expected_mc_size = kwargs.get("mc_size", 10000)
    assert criterion._mc_size == expected_mc_size
    assert criterion._batch_size == 2


def test_improvement_parallel_at_training_point(kriging_distribution):
    """Check that the improvement criteria take predefined values a training point."""
    criterion = EI(kriging_distribution, 2)
    criterion._compute_samples = lambda x: TypeError
    assert_almost_equal(criterion.evaluate(array([0.0, 0.0])), 0)


@pytest.mark.parametrize(
    ("cls", "input_value"),
    [
        (EI, array([0.123])),
        (UCB, array([0.123])),
        (Mean, array([0.123])),
    ],
)
def test_bad_parallel_regressor(algo_distribution, cls, input_value):
    """Check that parallelized criteria with non GP regressor lead to failure."""
    criterion = cls(algo_distribution, batch_size=2)

    with pytest.raises(
        NotImplementedError,
        match=re.escape(
            "Parallelization with batch_size > 1 is not yet implemented "
            "for regressors that are not based on a random process."
        ),
    ):
        criterion.func(input_value)



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Expected improvement."""

from __future__ import annotations

from gemseo_mlearning.active_learning.acquisition_criteria.maximum._expected_impovement import (  # noqa: E501
    ExpectedImprovement,
)
from gemseo_mlearning.active_learning.acquisition_criteria.maximum.base_maximum import (
    BaseMaximum,
)


class EI(ExpectedImprovement, BaseMaximum):
    r"""Expected improvement.

    This acquisition criterion is expressed as

    $$EI[x] = \mathbb{E}[\max(Y(x)-y_{\text{max}},0)]$$

    where $Y$ is the random process
    modelling the uncertainty of the surrogate model $\hat{f}$
    and $y_{\text{max}}$ is the maximum output value in the learning set.

    In the case of a Gaussian process regressor,
    it has an analytic expression:

    $$
    EI[x] = (\mathbb{E}[Y(x)] - y_{\text{max}})\Phi(t)
    + \mathbb{S}[Y(x)]\phi(t)
    $$

    where $\Phi$ and $\phi$ are respectively
    the cumulative and probability density functions
    of the standard normal distribution
    and $t=\frac{\mathbb{E}[Y(x)] - y_{\text{max}}}{\mathbb{S}[Y(x)]}$.

    For the acquisition of $q>1$ points at a time,
    the acquisition criterion changes to

    $$EI[x_1,\dots,x_q] = \mathbb{E}\left[\max_{1\leq i\leq q}\left(
    \max(Y(x_i)-y_{\text{max}},0)\right)\right]$$

    where the expectation is taken with respect to the distribution of
    the random vector $(Y(x_1),\dots,Y(x_q))$.
    There is no analytic expression
    and the acquisition is thus instead evaluated with crude Monte-Carlo.
    """



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Mixin for expected improvement."""

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING
from typing import ClassVar
from typing import Literal

from numpy import array
from numpy import atleast_2d
from numpy import dot
from numpy import max as np_max
from numpy import maximum
from numpy import mean
from numpy import nan_to_num
from scipy.stats import norm

if TYPE_CHECKING:
    from gemseo.typing import NumberArray

    from gemseo_mlearning.active_learning.distributions.regressor_distribution import (
        RegressorDistribution,
    )

warnings.filterwarnings("ignore")


class ExpectedImprovement:
    """A mixin for expected improvement."""

    _SIGN: ClassVar[Literal[-1, 1]] = 1
    """The sign."""

    def _compute(  # noqa: D102
        self, input_data: NumberArray
    ) -> NumberArray | float:
        # See Proposition 14, Jones et al, 1998
        improvement = (self._compute_mean(input_data) - self._qoi) * self._SIGN
        std = self._compute_standard_deviation(input_data)
        value = nan_to_num(improvement / std)
        return (
            improvement * norm.cdf(value) + std * norm.pdf(value)
        ) / self._scaling_factor

    def _compute_empirically(  # noqa: D102
        self, input_data: NumberArray
    ) -> NumberArray | float:
        self._regressor_distribution: RegressorDistribution
        ndim_is_two = input_data.ndim == 2
        input_data = atleast_2d(input_data)
        predictions = self._regressor_distribution.predict_members(input_data)
        weights = self._regressor_distribution.evaluate_weights(input_data)
        expected_improvement = maximum((predictions - self._qoi) * self._SIGN, 0.0)
        expected_improvement = array([
            dot(weights[:, index], expected_improvement[:, index, :])
            for index in range(expected_improvement.shape[1])
        ])
        if ndim_is_two:
            return expected_improvement / self._scaling_factor

        return expected_improvement[0] / self._scaling_factor

    def _compute_by_batch(  # noqa: D102
        self, q_input_values: NumberArray
    ) -> NumberArray | float:
        q_input_values = self._reshape_input_values(q_input_values)
        try:
            samples = self._compute_samples(
                input_data=q_input_values, n_samples=self._mc_size
            )[..., 0]
            improvement = (samples - self._qoi) * self._SIGN
            return mean(np_max(maximum(improvement, 0), axis=0)) / self._scaling_factor
        # distribution is not positive definite.
        except TypeError:
            return 0.0



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""A regression model from SMT."""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import ClassVar

from gemseo.mlearning.regression.algos.base_regressor import BaseRegressor
from numpy import concatenate
from numpy import double
from numpy import newaxis
from strenum import StrEnum

from gemseo_mlearning.regression.smt_regressor_settings import SMT_Regressor_Settings

if TYPE_CHECKING:
    from gemseo.typing import NumberArray

from gemseo_mlearning.regression.smt_regressor_settings import _NAMES_TO_CLASSES

SMTSurrogateModel = StrEnum("SurrogateModel", list(_NAMES_TO_CLASSES.keys()))
"""The class name of an SMT surrogate model."""


class SMTRegressor(BaseRegressor):
    """A regression model from SMT.

    !!! note

        [SMT](https://smt.readthedocs.io/) is an open-source Python package
        consisting of libraries of surrogate modeling methods,
        sampling methods, and benchmarking problems.
        [Read this page](https://smt.readthedocs.io/en/latest/_src_docs/surrogate_models.html)
        for the list of surrogate models and options.
    """

    Settings: ClassVar[type[SMT_Regressor_Settings]] = SMT_Regressor_Settings

    def _post_init(self):
        super()._post_init()
        model_class = _NAMES_TO_CLASSES[str(self._settings.model_class_name)]
        self.algo = model_class(**self._settings.parameters)

    def _fit(self, input_data: NumberArray, output_data: NumberArray) -> None:
        """
        Raises:
            ValueError: When the training dataset does not include gradient samples
                in the case of a gradient-enhanced surrogate model.
        """  # noqa: D205 D212 D415
        self.algo.set_training_values(input_data, output_data)
        surrogate_model_class_name = self.algo.__class__.__name__
        if not surrogate_model_class_name.startswith("GE"):
            self.algo.train()
            return

        # Gradient-enhanced (GE) surrogate models learn both outputs and gradients.
        try:
            get_view = self.learning_set.get_view
            gradient_group_name = self.learning_set.GRADIENT_GROUP
            jac_data = get_view(group_names=gradient_group_name).to_numpy()
        except KeyError as err:
            msg = (
                f"{surrogate_model_class_name} did not found gradient samples "
                "in the training dataset."
            )
            raise ValueError(msg) from err

        for i in range(self.input_dimension):
            self.algo.set_training_derivatives(
                input_data, jac_data[:, i :: self.input_dimension], i
            )

        self.algo.train()

    def _predict(self, input_data: NumberArray) -> NumberArray:
        return self.algo.predict_values(input_data.astype(double))

    def _predict_jacobian(self, input_data: NumberArray) -> NumberArray:
        return concatenate(
            tuple(
                self.algo.predict_derivatives(input_data, i)[..., newaxis]
                for i in range(input_data.shape[1])
            ),
            axis=2,
        )



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Contributors:
#    AUTHORS:
#       - Francois Gallard
"""Settings for the multi-start algorithm."""

from __future__ import annotations

from gemseo.algos.opt.base_optimizer_settings import BaseOptimizerSettings
from pydantic import Field
from pydantic import NonNegativeInt
from pydantic import PositiveInt
from smt.surrogate_models.surrogate_model import SurrogateModel  # noqa: TC002
from strenum import StrEnum


class AcquisitionCriterion(StrEnum):
    r"""An acquisition criterion.

    In the following, the training output values already used and the output and
    uncertainty predictions at a given input point $x$ are respectively denoted
    $\{y_1,\ldots,y_n\}$, $\mu(x)$ and $\sigma(x)$.
    """

    EI = "EI"
    r"""The expected improvement.

    The acquisition criterion is $\mathbb{E}[\max(\min(y_1,\dots,y_n)-Y,0]$ where $Y$ is
    a Gaussian random variable with mean $\mu(x)$ and standard deviation $\sigma(x)$.
    """

    LCB = "LCB"
    r"""The lower confidence bound.

    The acquisition criterion is $\mu(x)-3\sigma(x)$.
    """

    SBO = "SBO"
    r"""The surrogate-based optimization.

    The acquisition criterion is $\mu(x)$.
    """


class ParallelStrategy(StrEnum):
    r"""The strategy to set the outputs of the virtual points for parallel acquisition.

    In the following, the output variable, the training output values already used, and
    the output and uncertainty predictions at a given input point $x$ are respectively
    denoted $y$, $\{y_1,\ldots,y_n\}$, $\mu(x)$ and $\sigma(x)$.
    """

    CLmin = "CLmin"
    r"""The minimum constant liar.

    The output of the virtual point at $x$ is defined by $\min \{y_1,\ldots,y_n\}$.
    """

    KB = "KB"
    r"""The Kriging believer.

    The output of the virtual point at $x$ is defined by $\mu(x)$.
    """

    KBLB = "KBLB"
    r"""The Kriging believer lower bound.

    The output of the virtual point at $x$ is defined by $\mu(x)-3\sigma(x)$.
    """

    KBRand = "KBRand"
    r"""The Kriging believer random bound.

    The output of the virtual point at $x$ is defined by $\mu(x)+\kappa(x)\sigma(x)$
    where $\kappa(x)$ is the realization of a random variable distributed according to
    the standard normal distribution.
    """

    KBUB = "KBUB"
    r"""The Kriging believer upper bound.

    The output of the virtual point at $x$ is defined by $\mu(x)+3\sigma(x)$.
    """


class Surrogate(StrEnum):
    """A surrogate model."""

    GPX = "GPX"
    """Kriging based on the `egobox` library."""

    KRG = "KRG"
    """Kriging."""

    KPLS = "KPLS"
    """Kriging using partial least squares (PLS) to reduce the input dimension."""

    KPLSK = "KPLSK"
    """A variant of KPLS."""

    MGP = "MGP"
    """A marginal Gaussian process (MGP) regressor."""


class SMT_EGO_Settings(BaseOptimizerSettings):  # noqa: N801
    """The settings of the SMT's SBO algorithm."""

    criterion: AcquisitionCriterion = Field(
        default=AcquisitionCriterion.EI, description="The acquisition criterion."
    )

    enable_tunneling: bool = Field(
        default=False,
        description=(
            "Whether to enable the penalization of points "
            "that have been already evaluated in EI criterion."
        ),
    )

    n_doe: PositiveInt = Field(
        default=10,
        description=(
            "The number of points of the initial LHS DOE to train the surrogate."
        ),
    )

    n_max_optim: PositiveInt = Field(
        default=20,
        description="The maximum number of iterations for each sub-optimization.",
    )

    n_parallel: PositiveInt = Field(
        default=1,
        description=(
            "The number of points to be acquired in parallel; "
            "if `1`, acquire points sequentially."
        ),
    )

    n_start: PositiveInt = Field(
        default=50, description="The number of sub-optimizations."
    )

    qEI: ParallelStrategy = Field(  # noqa: N815
        default=ParallelStrategy.KBUB,
        description="The strategy to acquire points in parallel.",
    )

    random_state: NonNegativeInt = Field(
        default=1,
        description=(
            "The Numpy RandomState object or seed number which controls random draws."
        ),
    )

    surrogate: Surrogate | SurrogateModel = Field(
        default=Surrogate.KRG,
        description=(
            "The SMT Kriging-based surrogate model used internally; "
            "either an instance of the surrogate before training or its class name."
        ),
    )


# TODO: API: remove this alias.
SMTEGOSettings = SMT_EGO_Settings



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This work is licensed under a BSD 0-Clause License.
#
# Permission to use, copy, modify, and/or distribute this software
# for any purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
# WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL
# THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
# FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
# WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
"""# Batch acquisition."""

from __future__ import annotations

import matplotlib.pyplot as plt
from gemseo import configure
from gemseo import configure_logger
from gemseo import sample_disciplines
from gemseo.mlearning.regression.algos.ot_gpr import OTGaussianProcessRegressor
from gemseo.uncertainty.statistics.empirical_statistics import EmpiricalStatistics
from numpy import concatenate
from numpy import unique

from gemseo_mlearning.active_learning.active_learning_algo import ActiveLearningAlgo
from gemseo_mlearning.problems.rosenbrock.rosenbrock_discipline import (
    RosenbrockDiscipline,
)
from gemseo_mlearning.problems.rosenbrock.rosenbrock_space import RosenbrockSpace

# Update the configuration of |g| to speed up the script (use configure() with care)
configure(False, False, True, False, False, False, False)

configure_logger()

# %%
# The use of active learning methods
# dedicated to quantile estimation
# is illustrated in this example.
# More specifically,
# we aim to test here
# the impact of batch acquisition,
# that is to say when several points
# are added at each method iteration,
# on the active learning procedure.
# The function with the quantile of interest is
# the Rosenbrock function $f(x_1,x_2)=(1-x_1)^2+100(x_2-x_1^2)^2$:
discipline = RosenbrockDiscipline()
# %%
# with $x_1$ and $x_2$ uniformly distributed over $[-2,2]^2$:
uncertain_space = RosenbrockSpace()

# %%
# First,
# we create an initial training dataset using an optimal LHS including 10 samples:
learning_dataset = sample_disciplines(
    [discipline], uncertain_space, "y", algo_name="OT_OPT_LHS", n_samples=10
)

# %%
# and two identical initial
# Gaussian process regressors from OpenTURNS:
regressor_1 = OTGaussianProcessRegressor(learning_dataset, trend="quadratic")
regressor_2 = OTGaussianProcessRegressor(learning_dataset, trend="quadratic")

# %%
# Then,
# we build two active learning algorithms
# to test the impact of batch acquisition,
# that is to say when several points
# are added at each method iteration,
# on the active learning procedure.
# We consider one algorithm with
# samples added one by one
# also called sequential acquisition (default),
# and in the second one,
# four samples instead of one
# are added at each iteration (parallel acquisition),
# using argument `batch_size` set to 4.
# All other settings are put to
# their default values.
level = 0.35
active_learning_1 = ActiveLearningAlgo(
    "Quantile",
    uncertain_space,
    regressor_1,
    level=level,
    uncertain_space=uncertain_space,
)
active_learning_2 = ActiveLearningAlgo(
    "Quantile",
    uncertain_space,
    regressor_2,
    level=level,
    uncertain_space=uncertain_space,
    batch_size=4,
)
active_learning_1.acquire_new_points(discipline, 20)
active_learning_2.acquire_new_points(discipline, 20)


# %%
# To study the results,
# we extract first
# the data associated
# to the history
# of the quantity of interest
# for both active learning procedures
history_1 = active_learning_1.qoi_history
history_2 = active_learning_2.qoi_history
# and we compare them in a plot
plt.plot(
    history_1[0], concatenate(history_1[1]), marker="o", label="Sequential acquisition"
)
plt.plot(
    history_2[0], concatenate(history_2[1]), marker="o", label="Parallel acquisition"
)
plt.xlabel("Number of evaluations")
plt.ylabel("Quantile")
plt.legend()
plt.show()

# %%
# We can also
# compare the estimated quantile
# from the active learning procedure
# to the Monte Carlo estimate
# for both algorithms
dataset = sample_disciplines(
    [discipline], uncertain_space, "y", algo_name="OT_MONTE_CARLO", n_samples=1000
)
reference_quantile = EmpiricalStatistics(dataset, ["y"]).compute_quantile(level)

# %%
# Finally,
# for both active learning algorithms,
# we plot the training points,
# alongside the original model.

# Creation of the grid
# and estimation of the different quantities
n_test = 10
observations = sample_disciplines(
    [discipline], uncertain_space, "y", algo_name="OT_FULLFACT", n_samples=n_test**2
).values

# Plotting the contours of the Rosenbrock function
# alongside the learning points
plt.figure()
points_1 = active_learning_1.regressor.learning_set.to_numpy()
points_2 = active_learning_2.regressor.learning_set.to_numpy()
level_set = plt.contour(
    unique(observations[:, 0]),
    unique(observations[:, 1]),
    observations[:, 2].reshape(n_test, n_test),
    levels=[reference_quantile["y"]],
    colors="red",
)
plt.clabel(level_set, levels=[reference_quantile["y"]], fontsize=10, colors="red")
plt.annotate("True level set", (-0.2, 0.75), color="red")
plt.contour(
    unique(observations[:, 0]),
    unique(observations[:, 1]),
    observations[:, 2].reshape(n_test, n_test),
)
bar = plt.colorbar()
bar.set_label(r"$f(x_1,x_2)$")
plt.scatter(
    points_1[:, 0],
    points_1[:, 1],
    marker="*",
    label="Learning points from algo with sequential acquisition",
)
plt.scatter(
    points_2[:, 0],
    points_2[:, 1],
    marker="*",
    label="Learning points from algo with parallel acquisition",
)
plt.xlabel(r"$x_1$")
plt.ylabel(r"$x_2$")
plt.legend()
plt.show()



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""U-function."""

from __future__ import annotations

from typing import ClassVar

from gemseo_mlearning.active_learning.acquisition_criteria.level_set.u import U as _U
from gemseo_mlearning.active_learning.acquisition_criteria.quantile.base_quantile import (  # noqa: E501
    BaseQuantile,
)


class U(BaseQuantile):
    r"""The U-function.

    This acquisition criterion is expressed as

    $$U[x] = \frac{|y_{\alpha}-\mathbb{E}[Y(x)]|}{\mathbb{S}[Y(x)])}$$

    where $Y$ is the random process
    modelling the uncertainty of the surrogate model $\hat{f}$
    and $y_{\alpha}$ is the $\alpha$-quantile of the model output.
    For numerical purposes,
    the expression effectively minimized corresponds to its square root.

    For the acquisition of $q>1$ points at a time,
    the acquisition criterion changes to

    $$U[x_1,\dots,x_q] = \mathbb{E}\left[\min_{1\leq i\leq q}\left(
    \left(\frac{y_{\alpha}-Y(x_i)}{\mathbb{S}[Y(x_i)]}\right)^2\right)\right]$$

    where the expectation is taken with respect to the distribution of
    the random vector $(Y(x_1),\dots,Y(x_q))$.
    There is no analytic expression
    and the acquisition is thus instead evaluated with crude Monte-Carlo.
    """

    MAXIMIZE: ClassVar[bool] = False

    _LEVEL_SET_CLASS = _U



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This work is licensed under a BSD 0-Clause License.
#
# Permission to use, copy, modify, and/or distribute this software
# for any purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
# WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL
# THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
# FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
# WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
"""# SMT's surrogate model.

The [surrogate modeling toolbox (SMT)](../../../user_guide/regression/smt.md)
is an open-source Python package for surrogate modeling,
with a focus on derivatives.
The [SMTRegressor][gemseo_mlearning.regression.smt_regressor.SMTRegressor] class
allows you to use any SMT's surrogate model in your GEMSEO processes,
including the gradient-enhanced surrogate models
as long as your training dataset includes both output and gradient samples
as explained at the end of [this page](../../../user_guide/regression/smt.md).

In this example,
we will approximate the Rosenbrock function[@molga2005test]

$$f(x,y) = (1-x)^2 + 100(y-x^2)^2$$

over the domain $[-2,2]^2$.
"""

from __future__ import annotations

from gemseo import compute_doe
from gemseo import sample_disciplines
from gemseo.datasets.io_dataset import IODataset
from gemseo.mlearning.regression.quality.r2_measure import R2Measure
from gemseo.post.dataset.zvsxy import ZvsXY

from gemseo_mlearning.problems.rosenbrock.rosenbrock_discipline import (
    RosenbrockDiscipline,
)
from gemseo_mlearning.problems.rosenbrock.rosenbrock_space import RosenbrockSpace
from gemseo_mlearning.regression.smt_regressor import SMTRegressor

# %%
# First,
# we create the Rosenbrock discipline:
discipline = RosenbrockDiscipline()

# %%
# and the input space:
input_space = RosenbrockSpace()

# %%
# Then,
# we use an optimized Latin hypercube sampling (LHS) technique
# to generate 20 samples:
training_data = sample_disciplines(
    [discipline], input_space, "y", algo_name="OT_OPT_LHS", n_samples=20
)

# %%
# From this learning dataset,
# we train an [SMTRegressor][gemseo_mlearning.regression.smt_regressor.SMTRegressor]
# based on the
# [SMT's RBF surrogate model](https://smt.readthedocs.io/en/latest/_src_docs/surrogate_models/rbf.html)
# with the basis function scaling parameter `d_0` set to 2.0 (instead of 1.0):
surrogate_model = SMTRegressor(
    training_data, model_class_name="RBF", parameters={"d0": 2}
)
surrogate_model.learn()

# %%
# Finally,
# we assess its quality:
r2 = R2Measure(surrogate_model)
r2_l = r2.compute_learning_measure()[0]
r2_cv = r2.compute_cross_validation_measure()[0]
test_data = sample_disciplines(
    [discipline], input_space, "y", algo_name="OT_MONTE_CARLO", n_samples=1000
)
r2_t = r2.compute_test_measure(test_data)[0]
f"Learning R2: {r2_l}; cross-validation R2: {r2_cv}; test R2: {r2_t}"

# %%
# see how good it is with its R2 close to 1 on the test dataset,
# and plot its output over a 20x20 grid:
input_data = compute_doe(input_space, algo_name="PYDOE_FULLFACT", n_samples=400)
output_data = surrogate_model.predict(input_data)
predictions = IODataset()
predictions.add_input_group(input_data, variable_names=["x1", "x2"])
predictions.add_output_group(output_data, variable_names=["y"])

plot = ZvsXY(predictions, "x1", "x2", "y", other_datasets=(training_data,))
plot.color = "white"
plot.colormap = "viridis"
plot.execute(save=False, show=True)



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Tests for the surrogate-based optimization algorithms."""

from __future__ import annotations

import pickle
import re
from pathlib import Path

import pytest
from gemseo.algos.opt.factory import OptimizationLibraryFactory
from gemseo.problems.optimization.rastrigin import Rastrigin
from gemseo.problems.optimization.rosenbrock import Rosenbrock
from numpy import array
from numpy.testing import assert_almost_equal
from pandas.testing import assert_frame_equal

from gemseo_mlearning.algos.opt.sbo_settings import AcquisitionCriterion
from gemseo_mlearning.algos.opt.smt.ego_settings import SMT_EGO_Settings
from gemseo_mlearning.algos.opt.smt.ego_settings import SMTEGOSettings


def test_default_settings():
    """Check the default settings of the surrogate-based optimizer."""
    assert_almost_equal(
        OptimizationLibraryFactory()
        .execute(Rastrigin(), algo_name="SBO", max_iter=12)
        .f_opt,
        array([0.04]),
        decimal=2,
    )


@pytest.mark.parametrize("criterion", AcquisitionCriterion)
def test_criterion(criterion):
    """Check the surrogate-based optimizer with different criteria."""
    assert_almost_equal(
        OptimizationLibraryFactory()
        .execute(Rastrigin(), algo_name="SBO", max_iter=12, criterion=criterion)
        .f_opt,
        array([0.04]),
        decimal=2,
    )


@pytest.mark.parametrize("max_iter", [8, 10])
def test_inconsistent_max_iter(max_iter, regression_algorithm):
    """Check that max_iter must be strictly greater than doe_size."""
    with pytest.raises(
        ValueError,
        match=re.escape(
            f"max_iter ({max_iter}) must be "
            f"strictly greater than the initial DOE size (10)."
        ),
    ):
        OptimizationLibraryFactory().execute(
            Rastrigin(), algo_name="SBO", max_iter=max_iter
        )

    # Except if the regression algorithm is already built.
    OptimizationLibraryFactory().execute(
        Rastrigin(),
        algo_name="SBO",
        max_iter=max_iter,
        regression_algorithm=regression_algorithm,
    )


def test_save(regression_algorithm, tmp_wd):
    """Check that the  regression algorithm can be pickled."""
    file_path = Path("model.pkl")
    OptimizationLibraryFactory().execute(
        Rastrigin(),
        algo_name="SBO",
        max_iter=3,
        acquisition_algorithm="OT_MONTE_CARLO",
        acquisition_settings={"n_samples": 10},
        regression_algorithm=regression_algorithm,
        regression_file_path=file_path,
    )
    with file_path.open("rb") as file:
        model = pickle.load(file)

    assert_frame_equal(model.learning_set, regression_algorithm.learning_set)


def test_problem_counters(enable_function_statistics):
    """Check the counters attached to the optimization problem."""
    problem = Rosenbrock()
    OptimizationLibraryFactory().execute(
        problem,
        algo_name="SBO",
        max_iter=13,
        doe_size=10,
        acquisition_algorithm="OT_MONTE_CARLO",
        acquisition_settings={"n_samples": 100},
    )
    assert problem.evaluation_counter.maximum == 13
    assert problem.evaluation_counter.current == 13
    assert problem.objective.n_calls == 14


def test_alias():
    """Verify that SMTEGOSettings is an alias of SMT_EGO_Settings."""
    assert SMTEGOSettings == SMT_EGO_Settings



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""U-function."""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import ClassVar

from numpy import inf
from numpy import mean
from numpy import min as np_min
from numpy import nan_to_num
from numpy import square

from gemseo_mlearning.active_learning.acquisition_criteria.level_set.base_level_set import (  # noqa: E501
    BaseLevelSet,
)

if TYPE_CHECKING:
    from gemseo.typing import NumberArray


class U(BaseLevelSet):
    r"""The U-function.

    This acquisition criterion is expressed as

    $$U[x] = \mathbb{E}\left[\left(\frac{y-Y(x)}{\mathbb{S}[Y(x)]}
    \right)^2\right]$$

    where $y$ is the model output value characterizing the level set
    and $Y$ is the random process
    modelling the uncertainty of the surrogate model $\hat{f}$.
    It has an analytic expression:

    $$U[x] = \left(\frac{y-\mathbb{E}[Y(x)]}{\mathbb{S}[Y(x)]}\right)^2.$$

    For numerical purposes,
    the expression effectively minimized corresponds to its square root.

    For the acquisition of $q>1$ points at a time,
    the acquisition criterion changes to

    $$U[x_1,\dots,x_q] = \mathbb{E}\left[\min_{1\leq i \leq q}\left(
    \left(\frac{y-Y(x_i)}{\mathbb{S}[Y(x_i)]}\right)^2\right)\right]$$

    where the expectation is taken with respect to the distribution of
    the random vector $(Y(x_1),\dots,Y(x_q))$.
    There is no analytic expression
    and the acquisition is thus instead evaluated with crude Monte-Carlo.
    """

    MAXIMIZE: ClassVar[bool] = False

    def _compute(self, input_value: NumberArray) -> NumberArray | float:  # noqa: D102
        # See Proposition 12, Roy \& Notz, 2014
        return nan_to_num(
            abs(self._output_value - self._compute_mean(input_value))
            / self._compute_standard_deviation(input_value),
            nan=nan_to_num(inf),
        )

    def _compute_by_batch(self, q_input_values: NumberArray) -> NumberArray | float:  # noqa: D102
        q_input_values = self._reshape_input_values(q_input_values)
        try:
            samples = self._compute_samples(
                input_data=q_input_values, n_samples=self._mc_size
            )[..., 0]
            return nan_to_num(
                mean(
                    np_min(
                        square(self._output_value - samples.T)
                        / self._compute_variance(q_input_values),
                        axis=1,
                    )
                ),
                nan=nan_to_num(inf),
            )
        # distribution is not positive definite.
        except TypeError:
            return nan_to_num(inf)



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
from __future__ import annotations

import pytest
from numpy import array
from numpy.testing import assert_almost_equal

from gemseo_mlearning.problems.branin.functions import compute_gradient
from gemseo_mlearning.problems.branin.functions import compute_output


def test_compute_output() -> None:
    """Check the output of the Branin function."""
    assert compute_output(array([0.0, 1.0])) == pytest.approx(17.5, abs=0.1)


def test_compute_gradient() -> None:
    """Check the gradient of the Branin function."""
    assert_almost_equal(
        compute_gradient(array([0.0, 1.0])), array([-327.3, -65.6]), decimal=1
    )



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Output standard deviation."""

from __future__ import annotations

from typing import TYPE_CHECKING

from gemseo_mlearning.active_learning.acquisition_criteria.exploration.base_exploration import (  # noqa: E501
    BaseExploration,
)

if TYPE_CHECKING:
    from gemseo.typing import NumberArray


class StandardDeviation(BaseExploration):
    r"""Output standard deviation.

    This acquisition criterion is expressed as:

    $$\sigma[x] = \sqrt{\mathbb{E}[(Y(x)-\mathbb{E}[Y(x)])^2]}$$

    where $Y$ is the random process
    modelling the uncertainty of the surrogate model $\hat{f}$.
    """

    def _compute(self, input_value: NumberArray) -> NumberArray | float:  # noqa: D102
        return self._compute_standard_deviation(input_value) / self._scaling_factor



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This work is licensed under a BSD 0-Clause License.
#
# Permission to use, copy, modify, and/or distribute this software
# for any purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
# WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL
# THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
# FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
# WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
"""# GP regressor."""

from __future__ import annotations

import matplotlib.pyplot as plt
from gemseo import configure
from gemseo import configure_logger
from gemseo import sample_disciplines
from gemseo.mlearning.regression.algos.gpr import GaussianProcessRegressor
from gemseo.mlearning.regression.algos.ot_gpr import OTGaussianProcessRegressor
from gemseo.mlearning.regression.quality.r2_measure import R2Measure
from numpy import unique

from gemseo_mlearning.active_learning.active_learning_algo import ActiveLearningAlgo
from gemseo_mlearning.problems.rosenbrock.rosenbrock_discipline import (
    RosenbrockDiscipline,
)
from gemseo_mlearning.problems.rosenbrock.rosenbrock_space import RosenbrockSpace

# Update the configuration of |g| to speed up the script (use configure() with care)
configure(False, False, True, False, False, False, False)

configure_logger()

# %%
# The use of active learning methods
# dedicated to exploration
# is illustrated in this example.
# More specifically,
# we aim to test here
# the impact of the choice
# to test the impact of the choice
# of the GP on the active learning procedure,
# The function to approximate is
# the Rosenbrock function $f(x_1,x_2)=(1-x_1)^2+100(x_2-x_1^2)^2$:
discipline = RosenbrockDiscipline()
# %%
# with $x_1$ and $x_2$ belonging to $[-2,2]^2$:
input_space = RosenbrockSpace()

# %%
# First,
# we create an initial training dataset using an optimal LHS including 10 samples:
learning_dataset = sample_disciplines(
    [discipline], input_space, "y", algo_name="OT_OPT_LHS", n_samples=10
)

# %%
# and one Gaussian process regressor OpenTURNS:
regressor_1 = OTGaussianProcessRegressor(learning_dataset, trend="quadratic")
# %%
# and the other from scikit-learn:
regressor_2 = GaussianProcessRegressor(learning_dataset)


# %%
# Then,
# we build two active learning algorithms
# to test the impact of the choice
# of the GP on the active learning procedure,
# one from the OpenTURNS library
# and the second from scikit-learn.
# Those two are notably different,
# in particular,
# GPs from scikit-learn
# do not include trend modeling.
# All other settings are put to
# their default value.
active_learning_1 = ActiveLearningAlgo("Exploration", input_space, regressor_1)
active_learning_2 = ActiveLearningAlgo("Exploration", input_space, regressor_2)
active_learning_1.acquire_new_points(discipline, n_samples=20)
active_learning_2.acquire_new_points(discipline, n_samples=20)

# %%
# Finally,
# for both active learning algorithms,
# we plot the training points,
# alongside the original model.

# Creation of the grid
# and estimation of the different quantities
n_test = 10
observations = sample_disciplines(
    [discipline], input_space, "y", algo_name="OT_FULLFACT", n_samples=n_test**2
).values

# Plotting the contours of the Rosenbrock function
# alongside the learning points.
plt.figure()
plt.contour(
    unique(observations[:, 0]),
    unique(observations[:, 1]),
    observations[:, 2].reshape(n_test, n_test),
)
bar = plt.colorbar()
bar.set_label(r"$f(x_1,x_2)$")
points_1 = active_learning_1.regressor.learning_set.to_numpy()
points_2 = active_learning_2.regressor.learning_set.to_numpy()
plt.scatter(points_1[:, 0], points_1[:, 1], marker="*", label="OpenTURNS GP")
plt.scatter(points_2[:, 0], points_2[:, 1], marker="*", label="scikit-learn GP")
plt.xlabel(r"$x_1$")
plt.ylabel(r"$x_2$")
plt.legend()
plt.show()
# %%
# In this implementation,
# changing the GP regressor,
# does not seem to significantly alter
# the learning procedure,
# as most acquisition points correspond
# for both surrogates.
# Thus,
# they have a similar
# R2 validation metric:
dataset_test = sample_disciplines(
    [discipline], input_space, "y", algo_name="OT_OPT_LHS", n_samples=50
)
R2_1 = R2Measure(active_learning_1.regressor).compute_test_measure(dataset_test)
R2_2 = R2Measure(active_learning_2.regressor).compute_test_measure(dataset_test)
(R2_1[0], R2_2[0])



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
from __future__ import annotations

import pytest
from numpy import array
from numpy.testing import assert_equal

from gemseo_mlearning.problems.rosenbrock.rosenbrock_function import RosenbrockFunction
from gemseo_mlearning.problems.rosenbrock.rosenbrock_problem import RosenbrockProblem
from gemseo_mlearning.problems.rosenbrock.rosenbrock_space import RosenbrockSpace


@pytest.mark.parametrize("kwargs", [{}, {"use_uncertain_space": False}])
def test_rosenbrock_problem(kwargs) -> None:
    """Check the Rosenbrock problem."""
    problem = RosenbrockProblem(**kwargs)
    assert isinstance(problem.objective, RosenbrockFunction)

    input_space = problem.design_space
    if kwargs:
        assert input_space.dimension == 2
        assert input_space.variable_names == ["x1", "x2"]
        assert_equal(input_space.get_lower_bounds(), array([-2.0, -2.0]))
        assert_equal(input_space.get_upper_bounds(), array([2.0, 2.0]))
    else:
        assert isinstance(input_space, RosenbrockSpace)



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
from __future__ import annotations

import pytest
from numpy import array
from numpy.testing import assert_equal

from gemseo_mlearning.active_learning.acquisition_criteria.exploration.distance import (  # noqa: E501
    Distance,
)


@pytest.mark.parametrize(
    ("input_value", "expected"),
    [
        (array([0.5]), array([0.0])),
        (array([0.25]), array([1.0])),
        (array([0.125]), array([0.5])),
        (array([[0.125], [0.25]]), array([0.5, 1.0])),
    ],
)
def test_distance(algo_distribution, input_value, expected):
    """Check the Distance criterion."""
    criterion = Distance(algo_distribution)
    assert_equal(criterion.evaluate(input_value), expected)



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""The Branin function and its gradient."""

from __future__ import annotations

from typing import Final

from numpy import array
from numpy import cos
from numpy import ndarray
from numpy import pi
from numpy import sin

__A: Final[float] = 5.1 / (4 * pi**2)
__B: Final[float] = 5 / pi
__C: Final[float] = 10 - 10 / (8 * pi)


def compute_output(x: ndarray) -> float:
    """Compute the output of the Branin function.

    Args:
        x: The input values.

    Returns:
        The output value.
    """
    x0 = 15 * x[0] - 5
    return (15 * x[1] - __A * x0**2 + __B * x0 - 6) ** 2 + __C * cos(x0) + 10


def compute_gradient(x: ndarray) -> ndarray:
    """Compute the gradient of the Branin function.

    Args:
        x: The input values.

    Returns:
        The value of the gradient of the Branin function.
    """
    x0 = 15 * x[0] - 5
    tmp = 15 * x[1] - __A * x0**2 + __B * x0 - 6
    return array([-15 * (__C * sin(x0) + 2 * tmp * (2 * __A * x0 - __B)), 30 * tmp])



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""SMT optimizaton EGO tests."""

from __future__ import annotations

import contextlib
import re
from typing import TYPE_CHECKING
from unittest import mock

import pytest
from gemseo.problems.optimization.rosenbrock import Rosenbrock
from numpy import array
from numpy.testing import assert_allclose
from numpy.testing import assert_equal
from scipy.optimize import rosen
from smt.applications import EGO
from smt.surrogate_models import KRG
from smt.utils.design_space import DesignSpace
from smt.utils.design_space import DesignSpace as SMTDesignSpace

from gemseo_mlearning.algos.opt.smt._parallel_evaluator import ParallelEvaluator
from gemseo_mlearning.algos.opt.smt.ego_settings import AcquisitionCriterion
from gemseo_mlearning.algos.opt.smt.ego_settings import ParallelStrategy
from gemseo_mlearning.algos.opt.smt.ego_settings import SMT_EGO_Settings
from gemseo_mlearning.algos.opt.smt.ego_settings import Surrogate
from gemseo_mlearning.algos.opt.smt.smt_ego import SMTEGO

if TYPE_CHECKING:
    from gemseo.typing import RealArray


class ScaledRosenbrock:
    """The Rosenbrock function with normalized inputs."""

    def __init__(self):
        self.last_eval = None

    def __call__(self, x: RealArray) -> float:
        """Evaluate the Rosenbrock function with normalized input data.

        Args:
            x: The input value.

        Returns:
            The output value of the Rosenbrock function.
        """
        self.last_eval = rosen(x * 4 - 2)
        return self.last_eval


@pytest.mark.parametrize("criterion", AcquisitionCriterion)
@pytest.mark.parametrize("surrogate", Surrogate)
def test_criteria(criterion, surrogate):  # noqa: N803
    """Test EGO with different criteria."""
    if surrogate == Surrogate.MGP and criterion == AcquisitionCriterion.SBO:
        return

    optimization_problem = Rosenbrock()
    SMTEGO().execute(
        optimization_problem,
        n_start=2,
        n_doe=5,
        max_iter=7,
        n_max_optim=5,
        criterion=criterion,
        surrogate=surrogate,
    )
    last_eval = optimization_problem.database.last_item["rosen"]

    # Reference results using SMT directly.
    design_space = DesignSpace(array([[0.0, 1.0], [0.0, 1.0]]))
    surrogate = SMTEGO._SMTEGO__NAMES_TO_CLASSES[surrogate](design_space=design_space)
    ego = EGO(
        surrogate=surrogate,
        n_iter=1,
        n_doe=5,
        criterion=criterion,
        n_start=2,
        random_state=1,
        n_max_optim=2,
        evaluator=ParallelEvaluator(1),
    )
    scaled_rosenbrock = ScaledRosenbrock()
    ego.optimize(fun=scaled_rosenbrock)
    assert_allclose(scaled_rosenbrock.last_eval, last_eval, atol=1e-2)


@pytest.mark.parametrize(
    ("n_parallel", "x_opt", "f_opt"),
    [
        (1, array([-0.282595, 0.676449]), array([37.23683274])),
        (2, array([-0.736361, 1.068123]), array([30.670789])),
    ],
)
@pytest.mark.parametrize("normalize_design_space", [False, True])
def test_batch(n_parallel, x_opt, normalize_design_space, f_opt):  # noqa: N803
    """Test EGO w/wo batch acquisition and w/wo design space normalization."""
    optimization_problem = Rosenbrock()
    optimization_result = SMTEGO().execute(
        optimization_problem,
        settings_model=SMT_EGO_Settings(
            n_start=2,
            n_doe=5,
            max_iter=7,
            n_max_optim=5,
            n_parallel=n_parallel,
            normalize_design_space=normalize_design_space,
        ),
    )
    assert_allclose(optimization_result.x_opt, x_opt, atol=1e-5)
    assert_allclose(optimization_result.f_opt, f_opt, atol=1e-3)


@pytest.mark.parametrize("criterion", AcquisitionCriterion)
@pytest.mark.parametrize("qEI", ParallelStrategy)
def test_criteria_strategies(criterion, qEI):  # noqa: N803
    """Test EGO with different criteria and strategies.

    We mock the EGO class of SMT to check that the options are correctly passed.
    """
    optimization_problem = Rosenbrock()
    with (
        mock.patch.object(EGO, "__init__", return_value=None) as ego,
        contextlib.suppress(AttributeError),
    ):
        SMTEGO().execute(
            optimization_problem,
            n_start=2,
            n_doe=5,
            max_iter=7,
            n_max_optim=5,
            criterion=criterion,
            n_parallel=2,
            qEI=qEI,
            random_state=123,
        )

    assert ego.call_args.kwargs["n_start"] == 2
    assert ego.call_args.kwargs["n_doe"] == 5
    assert ego.call_args.kwargs["n_iter"] == 1
    assert ego.call_args.kwargs["n_max_optim"] == 5
    assert ego.call_args.kwargs["criterion"] == criterion
    assert ego.call_args.kwargs["n_parallel"] == 2
    assert ego.call_args.kwargs["qEI"] == qEI
    assert ego.call_args.kwargs["random_state"] == 123


@pytest.mark.parametrize(
    ("max_iter", "expected"),
    [
        (10, "max_iter must be strictly greater than n_doe+1 = 11; got 10."),
        (
            11,
            "max_iter must be strictly greater than n_doe+1 = 11; got 11.",
        ),
    ],
)
def test_max_iter(max_iter, expected):
    """Check that a ValueError is raised when max_iter is too small."""
    with pytest.raises(ValueError, match=re.escape(expected)):
        SMTEGO().execute(Rosenbrock(), max_iter=max_iter)


def test_from_existing_surrogate():
    """Test EGO from an existing surrogate."""
    result = SMTEGO().execute(
        Rosenbrock(),
        n_start=2,
        n_doe=5,
        max_iter=7,
        n_max_optim=5,
        surrogate="KRG",
    )
    x_opt = result.x_opt

    result = SMTEGO().execute(
        Rosenbrock(),
        n_start=2,
        n_doe=5,
        max_iter=7,
        n_max_optim=5,
        surrogate=KRG(
            # The design space is not [-2,2] but the unit hypercube,
            # because the optimization algorithm normalizes the design values.
            design_space=SMTDesignSpace(array([[0.0, 1.0], [0.0, 1.0]])),
            print_global=False,
        ),
    )

    assert_equal(result.x_opt, x_opt)
